/*  image.c
 *  
 *  Image processing functions of xwxapt application
 */  

/*
 *  xwxapt: An application to decode APT signals from
 *  weather satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or 
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "xwxapt.h"
#include "interface.h"
#include "support.h"

extern GtkWidget
  *xwxapt_top_window, /* xwxapt top window  */
  *gbl_text_scroller, /* Text view scroller */
  *gbl_level_scale,   /* Audio level scale  */
  *gbl_sync_scale,    /* Sync level scale   */
  *gbl_status_table,  /* Status widgets' table */
  *gbl_sync_icon,     /* Sync status indicator */
  *gbl_sync_label,    /* Sync status ind label */
  *gbl_apt_image;     /* Decoded weather image */

/* Text buffer for text view */
extern GtkTextBuffer *gbl_text_buffer;

/* Pixel buffer for display */
extern GdkPixbuf *gbl_pixbuf;

/* Pointer to pixbuf pixels */
extern guchar *gbl_pixel_buf;

/* Pixbuf rowstride and num of channels */
extern gint
  gbl_rowstride,
  gbl_n_channels;

extern int
  gbl_duration,   /* Duration of realtime operations */
  gbl_sat_type,   /* Weather satellite type number   */
  gbl_sync_ref,   /* Required position of sync train */
  gbl_buffer_idx; /* Index to audio samples buffer   */

/* Satellite type names */
extern char *gbl_sat_names[];

/* List of required sync position references */
extern int gbl_sync_refs[];

/* Pointer to samples file */
extern FILE *gbl_samples_fp;

/* Audio/DSP samples buffer */
unsigned char gbl_line_buffer[BUFFER_SIZE];

char
  gbl_samples_file[MAX_FILE_NAME+1], /* Sample file name */
  gbl_image_file  [MAX_FILE_NAME+1]; /* Image file name  */


/*------------------------------------------------------------------------*/

/*  Detect_NOAA_Sync()
 *
 *  Detects the Channel A sync of a NOAA APT signal.
 *  This is 7 cycles at 1040 Hz on the 2.4 KHz sub-carrier.
 */

int
Detect_NOAA_Sync( int search_limit )
{
  /* FIFO Buffer for ch A sync detection */
  static unsigned char sync_fifo[NOAA_FIFO_SIZE];

  int
    sample_a,    /* First of the pair of carrier samples  */
    sample_b,    /* Second of the pair of carrier samples */
    sync_cnt,    /* Counter used in detecting sync train position  */
    carrier_val, /* Instantaneous value of 2.4 Khz sub-carrier     */
    sync_val,    /* Value produced by sync detection algorithm     */
    sync_max,    /* Max value produced by sync detection algorithm */
    sync_ave,    /* Average value of carrier samples in sync fifo  */
    sync_pos,    /* Position in sample buffer where sync o/p peaks */
    peak_idx,    /* Index used in searching for peak in sync value */
    fifo_idx,    /* Index for sync detection fifo */
    idx;         /* Index for various loops etc.  */


  /* Move through buffer up to search_limit, looking   */
  /* for sync detector peak. On exit, the buffer index */
  /* should be at the end of A sync pulse sequence.    */

  sync_max = fifo_idx = 0;
  sync_pos = gbl_buffer_idx;

  for( peak_idx = 0; peak_idx < search_limit-1; peak_idx++ )
    {
      /* Read first of the pair of carrier samples */
      sample_a = gbl_line_buffer[gbl_buffer_idx++];

      /* Read second of the pair of carrier samples */
      sample_b = gbl_line_buffer[gbl_buffer_idx];

      /* This is the instantaneous magnitude of the 2.4 Khz  */
      /* subcarrier, calculated as the scalar magnitude of   */
      /* two consecutive audio samples at 90 deg phase angle */
      carrier_val = Scalar_Mag( sample_a, sample_b );

      /* Store 63 samples in fifo in cyclic manner */
      if( ++fifo_idx == NOAA_FIFO_SIZE )
	fifo_idx = 0;
      sync_fifo[fifo_idx] = carrier_val;

      /* Evaluate the average of the samples in fifo, to */
      /* be used as a zero reference in summation below. */
      sync_val = 0;
      for( idx = 0; idx < NOAA_FIFO_SIZE; idx++ )
	sync_val += sync_fifo[idx];
      sync_ave = sync_val/NOAA_FIFO_SIZE;

      /* Add and subtract alternate 4 sample groups in the   */
      /* sync accumulator, simulating synchronous detection. */
      /* Drop one sample between each pair of groups as      */
      /* there are ~9 carrier sample values per sync cycle.  */
      /* Their average value is used as a zero reference.    */
      sync_val = sync_cnt = 0;
      while( sync_cnt < NOAA_FIFO_SIZE )
	{
          /* Add 4 carrier samples into accumulator */
	  for( idx = 0; idx < 4; idx++ )
	    {
	      sync_val += sync_fifo[fifo_idx] - sync_ave;
	      if( ++fifo_idx == NOAA_FIFO_SIZE )
		fifo_idx = 0;
	      sync_cnt++;
	    }

	  /* Drop one sample between two 4-sample groups  */
	  /* This seems the easiest way to accomodate ~9  */
	  /* carrier samples per sync cycle of sync train */
	  if( ++fifo_idx == NOAA_FIFO_SIZE )
	    fifo_idx = 0;
	  sync_cnt++;

	  for( idx = 0; idx < 4; idx++ )
	    {
	      /* Subtract 4 carrier samples from accumulator */
	      sync_val -= sync_fifo[fifo_idx] - sync_ave;
	      if( ++fifo_idx == NOAA_FIFO_SIZE )
		fifo_idx = 0;
	      sync_cnt++;
	    }

	} /* End of while( sync_cnt < NOAA_FIFO_SIZE ) */

      /* Record position in buffer where sync detector peaks */
      if( sync_val > sync_max )
	{
	  sync_max = sync_val;
	  sync_pos = gbl_buffer_idx;
	}

    } /* End of for( peak_idx = 0; peak_idx < search_limit-1; peak_idx++ ) */

  if( isFlagClear(ACTION_PROCESS_FILE) )
    {
      /* Adjust sync detector scale */
      gtk_range_set_value( GTK_RANGE(gbl_sync_scale), (gdouble)sync_max );

      /* Setup buffer index to max sync position */
      /* if sync maximum is within the threshold */
      if( (sync_max > NOAA_SYNC_LOWER_THRESH) &&
	  (sync_max < NOAA_SYNC_UPPER_THRESH) )
	{
	  gbl_buffer_idx = sync_pos;

	  /* If sync position is in range, show 'Lock' */
	  if( (abs(gbl_buffer_idx - gbl_sync_ref) < 3) &&
	      isFlagClear(ACTION_SYNC_BUFFER) )
	    {
	      if( isFlagClear(ICON_SYNC_LOCK) )
		{
		  ClearFlag( ICON_FLAGS_ALL );
		  SetFlag( ICON_SYNC_LOCK );
		  Set_Sync_Icon( "gtk-yes" );
		  gtk_label_set_label( GTK_LABEL(gbl_sync_label), "Lock" );
		}
	    }
	  else /* Show 'Search' */
	    {
	      if( isFlagClear(ICON_SYNC_SEARCH) )
		{
		  ClearFlag( ICON_FLAGS_ALL );
		  SetFlag( ICON_SYNC_SEARCH );
		  Set_Sync_Icon( "gtk-dialog-question" );
		  gtk_label_set_label( GTK_LABEL(gbl_sync_label), "Search" );
		}

	    } /* if( abs(gbl_buffer_idx - gbl_sync_ref) < 3 ) */
	}
      else /* If sync level is out of range, show 'Unlock' */
	{
	  gbl_buffer_idx = 0;

	  if( isFlagClear(ICON_SYNC_UNLOCK) )
	    {
	      ClearFlag( ICON_FLAGS_ALL );
	      SetFlag( ICON_SYNC_UNLOCK );
	      Set_Sync_Icon( "gtk-no" );
	      gtk_label_set_label( GTK_LABEL(gbl_sync_label), "Unlock" );
	    }

	} /* if( (sync_max > NOAA_SYNC_LOWER_THRESH) && */

    } /* if( isFlagClear(ACTION_PROCESS_FILE) ) */
  else
    {
      /* Setup buffer index to max sync position */
      /* if sync maximum is within the threshold */
      if( (sync_max > NOAA_SYNC_LOWER_THRESH) &&
	  (sync_max < NOAA_SYNC_UPPER_THRESH) )
	gbl_buffer_idx = sync_pos;
      else
	gbl_buffer_idx = 0;
    }

  return(0);

} /* End of Detect_NOAA_Sync() */

/*------------------------------------------------------------------------*/

/*  Detect_Meteor_Sync()
 *
 *  Detects the sync train of a Meteor type APT signal.
 *  This is 5 cycles at 1200 Hz on the 2.4 KHz sub-carrier.
 */

int
Detect_Meteor_Sync( int search_limit )
{
  /* FIFO Buffer for sync detection */
  static unsigned char sync_fifo[METEOR_FIFO_SIZE];

  int
    sample_a,      /* First of the pair of carrier samples  */
    sample_b,      /* Second of the pair of carrier samples */
    sync_cnt,      /* Counter used in detecting sync train position  */
    carrier_val,   /* Instantaneous value of 2.4 Khz sub-carrier     */
    sync_val,      /* Value produced by sync detection algorithm     */
    sync_max,      /* Max value produced by sync detection algorithm */
    sync_ave,      /* Average value of carrier samples in sync fifo  */
    sync_pos,      /* Position in sample buffer where sync o/p peaks */
    peak_idx,      /* Index used in searching for peak in sync value */
    fifo_idx,      /* Index for sync detection fifo */
    idx;           /* Index for various loops etc.  */


  /* Move through buffer until sync detector  */
  /* peaks. On exit, the buffer index should  */
  /* be at the end of the sync pulse sequence */

  sync_max = fifo_idx = 0;
  sync_pos = gbl_buffer_idx;

  for( peak_idx = 0; peak_idx < search_limit-1 ; peak_idx++ )
    {
      /* Read first of the pair of carrier samples */
      sample_a = gbl_line_buffer[gbl_buffer_idx++];

      /* Read second of the pair of carrier samples */
      sample_b = gbl_line_buffer[gbl_buffer_idx];

      /* This is the instantaneous magnitude of the 2.4 Khz  */
      /* subcarrier, calculated as the scalar magnitude of   */
      /* two consecutive audio samples at 90 deg phase angle */
      carrier_val = Scalar_Mag( sample_a, sample_b );

      /* Store 40 samples in fifo in cyclic manner */
      if( ++fifo_idx == METEOR_FIFO_SIZE )
	fifo_idx = 0;
      sync_fifo[fifo_idx] = carrier_val;

      /* Evaluate the average of the samples in fifo, to */
      /* be used as a zero reference in summation below. */
      sync_val = 0;
      for( idx = 0; idx < METEOR_FIFO_SIZE; idx++ )
	sync_val += sync_fifo[idx];
      sync_ave = sync_val/METEOR_FIFO_SIZE;

      /* Subtract 5x3-sample groups and add 5x5-sample groups */
      /* to the sync accumulator alternately, simulating      */
      /* synchromous detection. The sync train appears to be  */
      /* 5 cycles at 1200 Hz with a 2/1 mark/space ratio.     */
      /* Their average value is used as a zero reference.     */
      sync_val = sync_cnt = 0;
      while( sync_cnt < METEOR_FIFO_SIZE )
	{
          /* Subtract 3 carrier samples from accumulator */
	  for( idx = 0; idx < 3; idx++ )
	    {
	      sync_val -= sync_fifo[fifo_idx] - sync_ave;
	      if( ++fifo_idx == METEOR_FIFO_SIZE )
		fifo_idx = 0;
	      sync_cnt++;
	    }

	  for( idx = 0; idx < 5; idx++ )
	    {
	      /* Add 5 carrier samples into accumulator */
	      sync_val += sync_fifo[fifo_idx] - sync_ave;
	      if( ++fifo_idx == METEOR_FIFO_SIZE )
		fifo_idx = 0;
	      sync_cnt++;
	    }

	} /* End of while( sync_cnt < METEOR_FIFO_SIZE ) */

      /* Record position in buffer where sync detector peaks */
      if( sync_val > sync_max)
	{
	  sync_max = sync_val;
	  sync_pos = gbl_buffer_idx;
	}

    } /* End of for( peak_idx = 0; peak_idx < search_limit-1; peak_idx++ ) */

  if( isFlagClear(ACTION_PROCESS_FILE) )
    {
      /* Adjust sync detector scale */
      gtk_range_set_value( GTK_RANGE(gbl_sync_scale), (gdouble)sync_max );

      /* Setup buffer index to max sync position */
      /* if sync maximum is within the threshold */
      if( (sync_max > METEOR_SYNC_LOWER_THRESH) &&
	  (sync_max < METEOR_SYNC_UPPER_THRESH) )
	{
	  gbl_buffer_idx = sync_pos;

	  /* If sync position is in range, show 'Lock' */
	  if( abs(gbl_buffer_idx - gbl_sync_ref) < 3 )
	    {
	      if( isFlagClear(ICON_SYNC_LOCK) )
		{
		  ClearFlag( ICON_FLAGS_ALL );
		  SetFlag( ICON_SYNC_LOCK );
		  Set_Sync_Icon( "gtk-yes" );
		  gtk_label_set_label( GTK_LABEL(gbl_sync_label), "Lock" );
		}
	    }
	  else /* Show 'Search' */
	    {
	      if( isFlagClear(ICON_SYNC_SEARCH) )
		{
		  ClearFlag( ICON_FLAGS_ALL );
		  SetFlag( ICON_SYNC_SEARCH );
		  Set_Sync_Icon( "gtk-dialog-question" );
		  gtk_label_set_label( GTK_LABEL(gbl_sync_label), "Search" );
		}

	    } /* if( abs(gbl_buffer_idx - gbl_sync_ref) < 3 ) */
	}
      else /* If sync level is out of range, show 'Unlock' */
	{
	  gbl_buffer_idx = 0;

	  if( isFlagClear(ICON_SYNC_UNLOCK) )
	    {
	      ClearFlag( ICON_FLAGS_ALL );
	      SetFlag( ICON_SYNC_UNLOCK );
	      Set_Sync_Icon( "gtk-no" );
	      gtk_label_set_label( GTK_LABEL(gbl_sync_label), "Unlock" );
	    }

	} /* if( (sync_max > METEOR_SYNC_LOWER_THRESH) && */

    } /* if( isFlagClear(ACTION_PROCESS_FILE) ) */
  else
    {
      /* Setup buffer index to max sync position */
      /* if sync maximum is within the threshold */
      if( (sync_max > METEOR_SYNC_LOWER_THRESH) &&
	  (sync_max < METEOR_SYNC_UPPER_THRESH) )
	gbl_buffer_idx = sync_pos;
      else
	gbl_buffer_idx = 0;
    }

  return(0);

} /* End of Detect_Meteor_Sync() */

/*------------------------------------------------------------------------*/

/*  Detect_Sync_Train()
 *
 *  Detects sync train position in line buffer
 *  depending on the type of satellite processed
 */

int
Detect_Sync_Train( int search_limit )
{
  switch( gbl_sat_type )
    {
    case NOAA   : return( Detect_NOAA_Sync(search_limit) );
    case METEOR : return( Detect_Meteor_Sync(search_limit) );
    }

  return(0);

} /* End of Detect_Sync_Train() */


/*------------------------------------------------------------------------*/

/*  NOAA_Image_Line()
 *
 *  Fills one line of one channel's image by
 *  reading samples from buffer and processing
 */

int
NOAA_Image_Line( unsigned char *image_line, int *max_gray_val )
{
  unsigned char
    sample_a, /* First of the pair of carrier samples  */
    sample_b; /* Second of the pair of carrier samples */

  int
    pixel_val, /* Un-normalized value of pixel */ 
    line_idx,  /* Index to image line buffer   */
    idx;       /* Index for loops etc */


  /* Process samples to make one line of image */
  for( line_idx = 0; line_idx < NOAA_IMAGE_WIDTH; line_idx++ )
    {
      pixel_val = 0;

      /*** Summate 3 carrier samples for each pixel value ***/
      for( idx = 0; idx < 3; idx++ )
	{
	  /* Read first of the pair of carrier samples */
	  sample_a = gbl_line_buffer[gbl_buffer_idx++];

	  /* Read second of the pair of carrier samples */
	  sample_b = gbl_line_buffer[gbl_buffer_idx];

	  /* This is the instantaneous magnitude of the 2.4 Khz  */
	  /* subcarrier, calculated as the scalar magnitude of   */
	  /* two consecutive audio samples at 90 deg phase angle */
	  pixel_val += (int)Scalar_Mag( sample_a, sample_b );

	} /* End of for( idx = 0; idx < 3; idx++ ) */

      /* Normalize pixel values to below 255 */
      image_line[line_idx] = (pixel_val * NOAA_SCALE_FACTOR) /
                                          NOAA_CARRIER_MAX;

      /* Record max gray value of line */
      if( image_line[line_idx] > *max_gray_val )
 	*max_gray_val = image_line[line_idx];

    } /* End of for( line_idx = 0; line_idx < NOAA_IMAGE_WIDTH; line_idx++ ) */

  return(0);

} /* End of NOAA_Image_Line() */

/*------------------------------------------------------------------------*/

/*  Meteor_Image_Line()
 *
 *  Fills one line of a Meteor APT image by
 *  reading samples from buffer and processing
 */

int
Meteor_Image_Line( unsigned char *image_line, int *max_gray_val )
{
  unsigned char
    sample_a, /* First of the pair of carrier samples  */
    sample_b; /* Second of the pair of carrier samples */

  int
    pixel_val, /* Un-normalized value of pixel */ 
    line_idx,  /* Index to image line buffer   */
    idx;       /* Index for loops etc */

  /* Process samples to make one line of image */
  for( line_idx = 0; line_idx < METEOR_IMAGE_WIDTH; line_idx++ )
    {
      pixel_val = 0;

      /*** Summate 5 carrier samples for each pixel value ***/
      for( idx = 0; idx < 5; idx++ )
	{
	  /* Read first of the pair of carrier samples */
	  sample_a = gbl_line_buffer[gbl_buffer_idx++];

	  /* Read second of the pair of carrier samples */
	  sample_b = gbl_line_buffer[gbl_buffer_idx];

	  /* This is the instantaneous magnitude of the 2.4 Khz  */
	  /* subcarrier, calculated as the scalar magnitude of   */
	  /* two consecutive audio samples at 90 deg phase angle */
	  pixel_val += (int)Scalar_Mag( sample_a, sample_b );

	} /* End of for( idx = 0; idx < 5; idx++ ) */

      /* Normalize pixel values to below 255 */
      image_line[line_idx] = (pixel_val * METEOR_SCALE_FACTOR) /
                                          METEOR_CARRIER_MAX;

      /* Record max gray value of line */
      if( image_line[line_idx] > *max_gray_val )
 	*max_gray_val = image_line[line_idx];

    } /* End of for( line_idx = 0; line_idx < image_width; line_idx++ ) */

  return(0);

} /* End of Meteor_Image_Line() */

/*------------------------------------------------------------------------*/

/*  Dsp_Process_Image()
 *
 *  Processes APT images direct from DSP audio samples.
 */ 

gboolean
Dsp_Process_Image( gpointer data )
{
  static int
    line_cnt; /* Number of APT image lines processed   */

    int
    carrier_val, /* Instantaneous value of 2.4Khz sub-carrier */
    carrier_max, /* Maximum value of 2.4KHz sub-carrier       */
    level_error; /* Error in 2.4KHz sub-carrier's max value   */ 

  /* Index for filling pixbuf */
  gint idx, idy;

  /* Message buffer */
  char message[350];

  /* NOAA images separator line */
  gint noaa_sep = NOAA_IMAGE_WIDTH * gbl_n_channels / 2;

  /* Pixbuf pixel pointer */
  guchar *pixel;

  /* First idle call */
  static gboolean first_call = TRUE;

  carrier_max = 0;

  /* Process APT images according to satellite type */
  switch( gbl_sat_type )
    {
    case NOAA : /* Decode NOAA APT images */
      {
	/* Image file names of NOAA ch A and B images */
	static char
	  A_image_file[MAX_FILE_NAME+20],
	  B_image_file[MAX_FILE_NAME+20];

	static int
	  A_max_gray_val,  /* Maximum gray value in channel A PGM image */
	  B_max_gray_val,  /* Maximum gray value in channel B PGM image */
	  A_image_height,  /* Height in lines of channel A PGM image    */
	  B_image_height,  /* Height in lines of channel B PGM image    */
	  A_image_size,    /* Current size in bytes of A image buffer   */
	  B_image_size;    /* Current size in bytes of B image buffer   */

	/* Averaged pixel value to enter in pixbuf */
	static int A_pixel_val;
	static int B_pixel_val;

	/* Buffers for the output images */
	static unsigned char *A_image_buffer;
	static unsigned char *B_image_buffer;

	static FILE
	  *A_image_fp, /* Channel A image file */
	  *B_image_fp; /* Channel B image file */


	if( first_call )
	  {
	    /* Fill pixbuf with background color */
	    gdk_pixbuf_fill( gbl_pixbuf, 0xaaaaaaff );

	    /* Process file name */
	    File_Name( gbl_image_file );

	    /* Prepare image file names */
	    strcpy( A_image_file, gbl_image_file );
	    strcpy( B_image_file, gbl_image_file );
	    strcat( A_image_file, "-NOAA-A.pgm" );
	    strcat( B_image_file, "-NOAA-B.pgm" );

	    /* Print some information */
	    sprintf( message, "Setting decoding duration to %d sec.",
		     gbl_duration );
	    Show_Message( message, "black" );

	    Show_Message( "Decoding images:", "black" );
	    sprintf( message, "%s", A_image_file );
	    Show_Message( message, "black" );
	    sprintf( message, "%s", B_image_file );
	    Show_Message( message, "black" );

	    /* Clear variables before decoding samples buffer */
	    A_image_height = B_image_height = 0;
	    A_image_size   = B_image_size   = 0;
	    A_max_gray_val = B_max_gray_val = 0;
	    A_image_buffer = B_image_buffer = NULL;
	    carrier_val    = line_cnt       = 0;

	    first_call = FALSE;

	  } /* if( first_call ) */

	/*** Process samples to make an image ***/

	/* Fill APT samples buffer */
	if( Dsp_Fill_Line_Buffer() == -1 )
	  {
	    Cleanup();
	    first_call = TRUE;
	    return( FALSE );
	  }

	/* Bypass sync and space/markers of ch A. */
	gbl_buffer_idx += NOAA_SYNC_CHA_SPACE;

	/* Re-allocate memory to image buffers, abort on error */
	A_image_buffer = realloc( A_image_buffer,
				  A_image_size + NOAA_IMAGE_WIDTH);

	if( A_image_buffer == NULL )
	  {
	    Show_Message( "A memory allocation request has failed - Quit!",
			  "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return(FALSE );
	  }

	B_image_buffer = realloc( B_image_buffer,
				  B_image_size + NOAA_IMAGE_WIDTH);

	if( B_image_buffer == NULL )
	  {
	    Show_Message( "A memory allocation request has failed - Quit!",
			  "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return(FALSE );
	  }

	/* Decode one image line */
	NOAA_Image_Line( A_image_buffer + A_image_size, &A_max_gray_val );
	A_image_height++;
	A_image_size = NOAA_IMAGE_WIDTH * A_image_height;

	/* Bypass ch A gray scale/ch B sync/ch B space/markers. */
	/* This is done by searching this zone for carrier max. */
	for( idx = 0; idx < NOAA_CHA_CHB_SPACE; idx++ )
	  {
	    carrier_val = Scalar_Mag(gbl_line_buffer[gbl_buffer_idx],
				     gbl_line_buffer[gbl_buffer_idx+1]);
	    gbl_buffer_idx++;

	    /* Find maximum carrier level */
	    if( carrier_val > carrier_max )
	      carrier_max = carrier_val;

	  } /* for( idx = 0; idx < NOAA_CHA_CHB_SPACE-1; idx++ ) */

	/* Find input audio level error and adjust scale */
	level_error = carrier_max - NOAA_CARRIER_MAX;
	gtk_range_set_value( GTK_RANGE(gbl_level_scale),
			     (gdouble)level_error );

	/* Decode an image line */
	NOAA_Image_Line( B_image_buffer + B_image_size, &B_max_gray_val );
	B_image_height++;
	B_image_size = NOAA_IMAGE_WIDTH * B_image_height;

	/* Draw images incrementally */
	idy = line_cnt/2;
	for( idx = 0; idx < NOAA_IMAGE_WIDTH/2; idx++ )
	  {
	    A_pixel_val = B_pixel_val = 0;

	    A_pixel_val += A_image_buffer[(2*idx)   +
					  (line_cnt * NOAA_IMAGE_WIDTH)];
	    A_pixel_val += A_image_buffer[(2*idx+1) +
					  (line_cnt * NOAA_IMAGE_WIDTH)];
	    B_pixel_val += B_image_buffer[(2*idx)   +
					  (line_cnt * NOAA_IMAGE_WIDTH)];
	    B_pixel_val += B_image_buffer[(2*idx+1) +
					  (line_cnt * NOAA_IMAGE_WIDTH)];

	    if( line_cnt & 0x01 )
	      {
		/* Find the A pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  (idx*gbl_n_channels);

		/* Assign A image R,G,B,A to grayscale value */
		A_pixel_val /= 2;
		pixel[0] = A_pixel_val;
		pixel[1] = A_pixel_val;
		pixel[2] = A_pixel_val;
		pixel[3] = 0xff;

		/* Find the B pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  ((idx+NOAA_IMAGE_WIDTH/2)*gbl_n_channels);

		/* Assign B image R,G,B,A to grayscale value */
		B_pixel_val /= 2;
		pixel[0] = B_pixel_val;
		pixel[1] = B_pixel_val;
		pixel[2] = B_pixel_val;
		pixel[3] = 0xff;

	      } /* if( line_cnt & 0x01 ) */
	    else
	      {
		/* Find the A pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  (idx*gbl_n_channels);

		/* Assign A image R,G,B,A to grayscale value */
		A_pixel_val /= 2;
		pixel[0] += A_pixel_val;
		pixel[1] += A_pixel_val;
		pixel[2] += A_pixel_val;
		pixel[3] += 0xff;

		/* Find the B pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  ((idx+NOAA_IMAGE_WIDTH/2)*gbl_n_channels);

		/* Assign B image R,G,B,A to grayscale value */
		B_pixel_val /= 2;
		pixel[0] += B_pixel_val;
		pixel[1] += B_pixel_val;
		pixel[2] += B_pixel_val;
		pixel[3] += 0xff;

	      } /* else */

	  } /* for( idx = 0; idx < NOAA_IMAGE_WIDTH/2; idx++ ) */

	/* Find the B pixel at 0,y and make it yellow */
	pixel = gbl_pixel_buf + (idy*gbl_rowstride) + noaa_sep;
	pixel[0] = 0xff;
	pixel[1] = 0xff;
	pixel[2] = 0xff;
	pixel[3] = 0xff;

	/* Draw image progressively to window */
	if( line_cnt & 0x01 )
	  gtk_widget_queue_draw( gbl_apt_image );

	/*** Continue if line count less than limit or action enabled ***/
	if( (++line_cnt/2 != gbl_duration) &&
	    isFlagSet(ACTION_PROCESS_DSP) )
	  return( TRUE );

	Show_Message( "Stopping Real Time image decoding.", "black" );
	Cleanup();

	/* Carry out histogram normalization if not suppressed */
	if( isFlagSet(IMAGE_NORMALIZE) )
	  {
	    Normalize( A_image_buffer, A_image_size );
	    Normalize( B_image_buffer, B_image_size );
	    A_max_gray_val = B_max_gray_val = 255;
	  }

	/* Rotate images 180 degrees if requested */
	if( isFlagSet(IMAGE_ROTATE) )
	  {
	    Rotate( A_image_buffer, A_image_size );
	    Rotate( B_image_buffer, B_image_size );
	  }

	/*** Processing finished, complete image files ***/

	/* Draw images */
	for( idy = 0; idy < line_cnt/2; idy++ )
	  {
	    /* Load image into pixbuff */
	    for( idx = 0; idx < NOAA_IMAGE_WIDTH/2; idx++ )
	      {
		A_pixel_val  = 0;
		A_pixel_val += A_image_buffer[(2*idx)   +
					      (2*idy)   * NOAA_IMAGE_WIDTH];
		A_pixel_val += A_image_buffer[(2*idx+1) +
					      (2*idy)   * NOAA_IMAGE_WIDTH];
		A_pixel_val += A_image_buffer[(2*idx)   +
					      (2*idy+1) * NOAA_IMAGE_WIDTH];
		A_pixel_val += A_image_buffer[(2*idx+1) +
					      (2*idy+1) * NOAA_IMAGE_WIDTH];
		A_pixel_val /= 4;

		/* Find the A pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  (idx*gbl_n_channels);

		/* Assign A image R,G,B,A to grayscale value */
		pixel[0] = A_pixel_val;
		pixel[1] = A_pixel_val;
		pixel[2] = A_pixel_val;
		pixel[3] = 0xff;

		/* Find the B pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  ((idx+NOAA_IMAGE_WIDTH/2)*gbl_n_channels);

		B_pixel_val  = 0;
		B_pixel_val += B_image_buffer[(2*idx)   +
					      (2*idy)   * NOAA_IMAGE_WIDTH];
		B_pixel_val += B_image_buffer[(2*idx+1) +
					      (2*idy)   * NOAA_IMAGE_WIDTH];
		B_pixel_val += B_image_buffer[(2*idx)   +
					      (2*idy+1) * NOAA_IMAGE_WIDTH];
		B_pixel_val += B_image_buffer[(2*idx+1) +
					      (2*idy+1) * NOAA_IMAGE_WIDTH];
		B_pixel_val /= 4;

		/* Assign B image R,G,B,A to grayscale value */
		pixel[0] = B_pixel_val;
		pixel[1] = B_pixel_val;
		pixel[2] = B_pixel_val;
		pixel[3] = 0xff;
	      }

	    /* Find the B pixel at 0,y and make it yellow */
	    pixel = gbl_pixel_buf + (idy*gbl_rowstride) + noaa_sep;
	    pixel[0] = 0xff;
	    pixel[1] = 0xff;
	    pixel[2] = 0xff;
	    pixel[3] = 0xff;

	  } /* for( idy = 0; idx < gbl_duration; idx++ ) */

	/* Draw image to window */
	gtk_widget_queue_draw( gbl_apt_image );

	/* Open Channel A image file */
	A_image_fp = fopen( A_image_file, "w+" );
	if( A_image_fp == NULL )
	  {
	    perror( A_image_file );
	    sprintf( message, "Failed to open file for writing:\n%s",
		     A_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return( FALSE );
	  }
 
	/* Open Channel B image file */
	B_image_fp = fopen( B_image_file, "w+" );
	if( B_image_fp == NULL )
	  {
	    perror( B_image_file );
	    sprintf( message, "Failed to open file for writing:\n%s",
		     B_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return( FALSE );
	  }

	/* Write header in Ch-A output PGM files */
	fprintf( A_image_fp, "P5\n" );
	fprintf( A_image_fp, "%d %d\n", NOAA_IMAGE_WIDTH, A_image_height );
	fprintf( A_image_fp, "%d\n", A_max_gray_val );

	/* Write Ch-A image buffer to file, abort on error */
	if( fwrite(A_image_buffer, 1, A_image_size, A_image_fp)
	    != A_image_size )
	  {
	    sprintf( message, "Error writing image to file:\n%s",
		     A_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return( FALSE );
	  }

	/* Write header in Ch-B output PGM file */
	fprintf( B_image_fp, "P5\n" );
	fprintf( B_image_fp, "%d %d\n", NOAA_IMAGE_WIDTH, B_image_height );
	fprintf( B_image_fp, "%d\n", B_max_gray_val );

	/* Write Ch-B image buffer to file, abort on error */
	if( fwrite(B_image_buffer, 1, B_image_size, B_image_fp)
	    != B_image_size )
	  {
	    sprintf( message, "Error writing image to file:\n%s",
		     B_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return( FALSE );
	  }

	/* Free buffers and close image files */
	Show_Message( "Closing image files:", "black" );
	sprintf( message, "%s", A_image_file );
	Show_Message( message, "black" );
	sprintf( message, "%s", B_image_file );
	Show_Message( message, "black" );

	free( A_image_buffer );
	free( B_image_buffer );
	fclose( A_image_fp );
	fclose( B_image_fp );

      } /* End of case NOAA : */

      break;

    case METEOR : /* Decode Meteor APT images */
      {
	/* Image file name of Meteor image */
	static char Met_image_file[MAX_FILE_NAME+20];

	static int
	  max_gray_val,     /* Maximum gray value in Meteor APT image */
	  Met_image_height, /* Height in lines of Meteor APT image    */
	  Met_image_size;   /* Current size in bytes of Meteor image  */

	/* Averaged pixel value to enter in pixbuf */
	static int pixel_val = 0;

	/* Buffer for the output image */
	static unsigned char *Met_image_buffer;

	/* Meteor APT image file pointer */
	static FILE *Met_image_fp;


	if( first_call )
	  {
	    /* Fill pixbuf with background color */
	    gdk_pixbuf_fill( gbl_pixbuf, 0xaaaaaaff );

	    /* Process file name */
	    File_Name( gbl_image_file );

	    /* Prepare image file name */
	    strcpy( Met_image_file, gbl_image_file );
	    strcat( Met_image_file, "-Meteor.pgm" );

	    /* Print some information */
	    if( gbl_duration )
	      {
		sprintf( message, "Setting decoding duration to %d sec.",
			 gbl_duration );
		Show_Message( message, "black" );
	      }

	    Show_Message( "Decoding image:", "black" );
	    sprintf( message, "%s", Met_image_file );
	    Show_Message( message, "black" );

	    /* Clear variables before decoding sample buffer */
	    Met_image_height = 0;
	    max_gray_val     = 0;
	    Met_image_size   = 0;
	    Met_image_buffer = NULL;
	    carrier_val      = 0;
	    line_cnt         = 0;

	    first_call = FALSE;

	  } /* if( first_call ) */

	/*** Process samples to make an image ***/

	/* Fill APT samples buffer */
	if( Dsp_Fill_Line_Buffer() == -1 )
	  {
	    Cleanup();
	    first_call = TRUE;
	    return( FALSE );
	  }

	/* Bypass sync and space/markers of image */
	/* This is done by searching this zone for carrier max. */
	for( idx = 0; idx < METEOR_SYNC_IMAGE_SPACE; idx++ )
	  {
	    carrier_val = Scalar_Mag(gbl_line_buffer[gbl_buffer_idx],
				     gbl_line_buffer[gbl_buffer_idx+1]);
	    gbl_buffer_idx++;

	    /* Find maximum carrier level */
	    if( carrier_val > carrier_max )
	      carrier_max = carrier_val;

	  } /* for( idx = 0; idx < NOAA_CHA_CHB_SPACE-1; idx++ ) */

	/* Find input audio level error and adjust scale */
	level_error = carrier_max - METEOR_CARRIER_MAX + 20;
	gtk_range_set_value( GTK_RANGE(gbl_level_scale), (gdouble)level_error );

	/* Re-allocate memory to image buffer, abort on error */
	Met_image_buffer = realloc( Met_image_buffer,
				    Met_image_size + METEOR_IMAGE_WIDTH);

	if( Met_image_buffer == NULL )
	  {
	    Show_Message( "A memory allocation request has failed - Quit!",
			  "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return( FALSE );
	  }

	/* Decode an image line */
	Meteor_Image_Line( Met_image_buffer + Met_image_size, &max_gray_val );
	Met_image_height++;
	Met_image_size = METEOR_IMAGE_WIDTH * Met_image_height;

	/* Draw images incrementally */
	idy = line_cnt/2;
	for( idx = 0; idx < METEOR_IMAGE_WIDTH/2; idx++ )
	  {
	    pixel_val = 0;

	    pixel_val += Met_image_buffer[(2*idx)   +
					  (line_cnt * METEOR_IMAGE_WIDTH)];
	    pixel_val += Met_image_buffer[(2*idx+1) +
					  (line_cnt * METEOR_IMAGE_WIDTH)];

	    if( line_cnt & 0x01 )
	      {
		/* Find the pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  (idx*gbl_n_channels);

		/* Assign image R,G,B,A to grayscale value */
		pixel_val /= 2;
		pixel[0] = pixel_val;
		pixel[1] = pixel_val;
		pixel[2] = pixel_val;
		pixel[3] = 0xff;

	      } /* if( line_cnt & 0x01 ) */
	    else
	      {
		/* Find the pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  (idx*gbl_n_channels);

		/* Assign image R,G,B,A to grayscale value */
		pixel_val /= 2;
		pixel[0] += pixel_val;
		pixel[1] += pixel_val;
		pixel[2] += pixel_val;
		pixel[3] += 0xff;

	      } /* else */

	  } /* for( idx = 0; idx < METEOR_IMAGE_WIDTH/2; idx++ ) */

	/* Progressively draw image to window */
	if( line_cnt & 0x01 )
	  gtk_widget_queue_draw( gbl_apt_image );

	/*** Continue if line count less than limit or action enabled ***/
	if( (++line_cnt/2 != gbl_duration) &&
	    isFlagSet(ACTION_PROCESS_DSP) )
	  return( TRUE );

	Show_Message( "Stopping Real Time image decoding.", "black" );
	Cleanup();

	/* Carry out histogram normalization if not suppressed */
	if( isFlagSet(IMAGE_NORMALIZE) )
	  {
	    Normalize( Met_image_buffer, Met_image_size );
	    max_gray_val = 255;
	  }

	/* Rotate images 180 degrees if requested */
	if( isFlagSet(IMAGE_ROTATE) )
	  Rotate( Met_image_buffer, Met_image_size );

	/*** Processing finished, complete image files ***/

	/* Draw images */
	for( idy = 0; idy < line_cnt/2; idy++ )
	  {
	    /* Load image into pixbuff */
	    for( idx = 0; idx < METEOR_IMAGE_WIDTH/2; idx++ )
	      {
		pixel_val  = 0;
		pixel_val += Met_image_buffer[(2*idx)   +
					      (2*idy)   * METEOR_IMAGE_WIDTH];
		pixel_val += Met_image_buffer[(2*idx+1) +
					      (2*idy)   * METEOR_IMAGE_WIDTH];
		pixel_val += Met_image_buffer[(2*idx)   +
					      (2*idy+1) * METEOR_IMAGE_WIDTH];
		pixel_val += Met_image_buffer[(2*idx+1) +
					      (2*idy+1) * METEOR_IMAGE_WIDTH];

		pixel_val /= 4;

		/* Find the pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  (idx*gbl_n_channels);

		/* Assign image R,G,B,A to grayscale value */
		pixel[0] = pixel_val;
		pixel[1] = pixel_val;
		pixel[2] = pixel_val;
		pixel[3] = 0xff;

	      } /* for( idx = 0; idx < METEOR_IMAGE_WIDTH/2; idx++ ) */

	  } /* for( idy = 0; idy < gbl_duration; idy++ ) */

	/* Draw finished image to window */
	gtk_widget_queue_draw( gbl_apt_image );

	/* Open Meteor image file */
	Met_image_fp = fopen( Met_image_file, "w+" );
	if( Met_image_fp == NULL )
	  {
	    perror( Met_image_file );
	    sprintf( message, "Failed to open file for writing:\n%s",
		     Met_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return( FALSE );
	  }

	/* Write header of Meteor output PGM files */
	fprintf( Met_image_fp, "P5\n" );
	fprintf( Met_image_fp, "%d %d\n",
		 METEOR_IMAGE_WIDTH, Met_image_height );
	fprintf( Met_image_fp, "%d\n", max_gray_val );

	/* Write image buffer to file, abort on error */ 
	if( (idx = fwrite(Met_image_buffer, 1, Met_image_size, Met_image_fp)) !=
	    Met_image_size )
	  {
	    sprintf( message, "Error writing image to file:\n%s",
		     Met_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    first_call = TRUE;
	    return( FALSE );
	  }

	/* Free image buffer and close image file */
	Show_Message( "Closing image file:", "black" );
	sprintf( message, "%s", Met_image_file );
	Show_Message( message, "black" );

	free( Met_image_buffer );
	fclose( Met_image_fp );

      } /* End of case METEOR : */

    } /* End of switch( gbl_sat_type ) */

  /* Enable re-initialization */
  first_call = TRUE;
  Cleanup();

  return( FALSE );

} /* End of Dsp_Process_Image() */

/*------------------------------------------------------------------------*/

/*  File_Process_Image()
 *
 *  Processes APT images from audio samples file.
 */ 

gboolean
File_Process_Image( gpointer data )
{
  /* Message buffer */
  char message[350];

  /* File suffix to be removed from file name */
  char suffix[20];

  /* Position of suffix in file name */
  char *suffix_pos;

  /* x,y index for filling pixbuf */
  int idx, idy;

  /* NOAA images separator */
  gint noaa_sep = NOAA_IMAGE_WIDTH * gbl_n_channels / 2;

  /* Pixbuf pixel pointer */
  guchar *pixel;


  /* Open APT audio samples file */
  sprintf( message, "Opening recorded samples file:\n%s",
	   gbl_samples_file );
  Show_Message( message, "black" );

  if( (gbl_samples_fp = fopen(gbl_samples_file, "r")) == NULL )
    {
      perror( gbl_samples_file );
      sprintf( message, "Cannot open file or directory:\n%s",
	       gbl_samples_file );
      Show_Message( message, "red" );
      Error_Dialog();
      return( FALSE );
    }

  /* Read satellite type from file, abort on error */
  if( fread(&gbl_sat_type, 4, 1, gbl_samples_fp) == -1 )
    {
      perror( "xwxapt: fread()" );
      sprintf( message, "Failed to read satellite type from file:\n%s",
	       gbl_samples_file );
      Show_Message( message, "red" );
      Error_Dialog();
      return( FALSE );
    }

  if( (gbl_sat_type != 0) &&
      (gbl_sat_type != 1) )
    {
      sprintf( message, "Incorrect satellite type in samples file:\n%s",
	       gbl_samples_file );
      Show_Message( message, "red" );
      Error_Dialog();
      return( FALSE );
    }

  /* Read sync position reference */
  gbl_sync_ref = gbl_sync_refs[ gbl_sat_type ];

  /* Creat the suffix to be removed from sample file name. */
  /* This is to remove the suffix <sat_name>.bin which is  */
  /* added automatically by xwxapt to samples file name.   */
  suffix[0] = '-';
  suffix[1] = '\0';
  strcat( suffix, gbl_sat_names[gbl_sat_type] );
  strcat( suffix, ".bin" );

  /* Find position of suffix in file name */
  suffix_pos = strstr( gbl_samples_file, suffix );

  /* Cut out suffix if it exists */ 
  if( suffix_pos != NULL )
    suffix_pos[0] = '\0';

  /* Process APT images according to satellite type */
  switch( gbl_sat_type )
    {
    case NOAA : /* Decode NOAA APT images */
      {
	/* Image file names of NOAA ch A and B images */
	char
	  A_image_file[MAX_FILE_NAME+20],
	  B_image_file[MAX_FILE_NAME+20];

	int
	  line_cnt,        /* Number of APT image lines processed       */
	  carrier_val,     /* Instantaneous value of 2.4 Khz subcarrier */
	  A_max_gray_val,  /* Maximum gray value in channel A PGM image */
	  B_max_gray_val,  /* Maximum gray value in channel B PGM image */
	  A_image_height,  /* Height in lines of channel A PGM image    */
	  B_image_height,  /* Height in lines of channel B PGM image    */
	  A_image_size,    /* Current size in bytes of A image buffer   */
	  B_image_size;    /* Current size in bytes of B image buffer   */

	/* Averaged pixel value to enter in pixbuf */
	int A_pixel_val, B_pixel_val;

	/* Buffers for the output images */
	unsigned char *A_image_buffer;
	unsigned char *B_image_buffer;

	FILE
	  *A_image_fp,   /* Channel A image file  */
	  *B_image_fp;   /* Channel B image file  */

	/* Process file name */
	if( strlen(gbl_image_file) == 0 )
	  strcpy( gbl_image_file, gbl_samples_file );
	else
	  File_Name( gbl_samples_file );

	/* Prepare image file names */
	strcpy( A_image_file, gbl_image_file );
	strcpy( B_image_file, gbl_image_file );
	strcat( A_image_file, "-NOAA-A.pgm" );
	strcat( B_image_file, "-NOAA-B.pgm" );

	/* Print some information */
	Show_Message( "Decoding images:", "black" );
	sprintf( message, "%s", A_image_file );
	Show_Message( message, "black" );
	sprintf( message, "%s", B_image_file );
	Show_Message( message, "black" );

	/* Clear variables before decoding samples buffer */
	A_image_height = B_image_height = 0;
	A_image_size   = B_image_size   = 0;
	A_max_gray_val = B_max_gray_val = 0;
	A_image_buffer = B_image_buffer = NULL;
	carrier_val    = line_cnt       = 0;

	/* Process samples from file to make an image */
	while( File_Fill_Buffer() != -1 )
	  {
	    /* Bypass A sync and space/markers of ch A. */
	    gbl_buffer_idx += NOAA_SYNC_CHA_SPACE;

	    /* Re-allocate memory to image buffers, abort on error */
	    A_image_buffer = realloc( A_image_buffer,
				      A_image_size + NOAA_IMAGE_WIDTH);

	    if( A_image_buffer == NULL )
	      {
		Show_Message( "A memory allocation request has failed - Quit!", 
			      "red" );
		Error_Dialog();
		return( FALSE );
	      }

	    B_image_buffer = realloc( B_image_buffer,
				      B_image_size + NOAA_IMAGE_WIDTH);

	    if( B_image_buffer == NULL )
	      {
		Show_Message( "A memory allocation request has failed - Quit!",
			      "red" );
		Error_Dialog();
		return( FALSE );
	      }

	    /* Decode an image line */
	    NOAA_Image_Line( A_image_buffer + A_image_size, &A_max_gray_val );
	    A_image_height++;
	    A_image_size = NOAA_IMAGE_WIDTH * A_image_height;

	    /* Bypass ch A gray scale/ch B sync/ch B space/markers */
	    gbl_buffer_idx += NOAA_CHA_CHB_SPACE;

	    /* Decode an image line */
	    NOAA_Image_Line( B_image_buffer + B_image_size, &B_max_gray_val );
	    B_image_height++;
	    B_image_size = NOAA_IMAGE_WIDTH * B_image_height;

	    line_cnt++;

	  } /* End of while( File_Fill_Buffer() != -1 ) */

	/* Carry out histogram normalization if not suppressed */
	if( isFlagSet(IMAGE_NORMALIZE) )
	  {
	    Normalize( A_image_buffer, A_image_size );
	    Normalize( B_image_buffer, B_image_size );
	    A_max_gray_val = B_max_gray_val = 255;
	  }

	/* Rotate images 180 degrees if requested */
	if( isFlagSet(IMAGE_ROTATE ) )
	  {
	    Rotate( A_image_buffer, A_image_size );
	    Rotate( B_image_buffer, B_image_size );
	  }

	/*** Processing finished, complete image files ***/

	/* Resize window as required */
	gbl_duration = line_cnt / 2;
	Initialize_Top_window();

	/* Draw images */
	for( idy = 0; idy < gbl_duration; idy++ )
	  {
	    /* Load image into pixbuff */
	    for( idx = 0; idx < NOAA_IMAGE_WIDTH/2; idx++ )
	      {
		A_pixel_val  = 0;
		A_pixel_val += A_image_buffer[(2*idx)   +
					      (2*idy)   * NOAA_IMAGE_WIDTH];
		A_pixel_val += A_image_buffer[(2*idx+1) +
					      (2*idy)   * NOAA_IMAGE_WIDTH];
		A_pixel_val += A_image_buffer[(2*idx)   +
					      (2*idy+1) * NOAA_IMAGE_WIDTH];
		A_pixel_val += A_image_buffer[(2*idx+1) +
					      (2*idy+1) * NOAA_IMAGE_WIDTH];

		A_pixel_val /= 4;

		/* Find the A pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  (idx*gbl_n_channels);

		/* Assign A image R,G,B,A to grayscale value */
		pixel[0] = A_pixel_val;
		pixel[1] = A_pixel_val;
		pixel[2] = A_pixel_val;
		pixel[3] = 0xff;

		/* Find the B pixel at x,y */
		pixel = gbl_pixel_buf +
		  (idy*gbl_rowstride) +
		  ((idx+NOAA_IMAGE_WIDTH/2)*gbl_n_channels);

		B_pixel_val  = 0;
		B_pixel_val += B_image_buffer[(2*idx)   +
					      (2*idy)   * NOAA_IMAGE_WIDTH];
		B_pixel_val += B_image_buffer[(2*idx+1) +
					      (2*idy)   * NOAA_IMAGE_WIDTH];
		B_pixel_val += B_image_buffer[(2*idx)   +
					      (2*idy+1) * NOAA_IMAGE_WIDTH];
		B_pixel_val += B_image_buffer[(2*idx+1) +
					      (2*idy+1) * NOAA_IMAGE_WIDTH];

		B_pixel_val /= 4;

		/* Assign B image R,G,B,A to grayscale value */
		pixel[0] = B_pixel_val;
		pixel[1] = B_pixel_val;
		pixel[2] = B_pixel_val;
		pixel[3] = 0xff;
	      }

	    /* Find the B pixel at 0,y and make it yellow */
	    pixel = gbl_pixel_buf + (idy*gbl_rowstride) + noaa_sep;

	    pixel[0] = 0xff;
	    pixel[1] = 0xff;
	    pixel[2] = 0xff;
	    pixel[3] = 0xff;

	  } /* for( idx = 0; idx < NOAA_IMAGE_WIDTH; idx++ ) */

	/* Draw finished image */
	gtk_widget_queue_draw( gbl_apt_image );

	/* Open Channel A image file */
	A_image_fp = fopen( A_image_file, "w+" );
	if( A_image_fp == NULL )
	  {
	    perror( A_image_file );
	    sprintf( message, "Failed to open file for writing:\n%s",
		     A_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    return( FALSE );
	  }

	/* Open Channel B image file */
	B_image_fp  = fopen( B_image_file, "w+" );
	if( B_image_fp == NULL )
	  {
	    perror( B_image_file );
	    sprintf( message, "Failed to open file for writing:\n%s",
		     B_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    return( FALSE );
	  }

	/* Write header in Ch-A output PGM files */
	fprintf( A_image_fp, "P5\n" );
	fprintf( A_image_fp, "%d %d\n", NOAA_IMAGE_WIDTH, A_image_height );
	fprintf( A_image_fp, "%d\n", A_max_gray_val );

	/* Write Ch-A image buffer to file, abort on error */
	if( fwrite(A_image_buffer, 1, A_image_size, A_image_fp) != A_image_size )
	  {
	    sprintf( message, "Error writing image to file:\n%s",
		     A_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    return( FALSE );
	  }

	/* Write header in Ch-B output PGM file */
	fprintf( B_image_fp, "P5\n" );
	fprintf( B_image_fp, "%d %d\n", NOAA_IMAGE_WIDTH, B_image_height );
	fprintf( B_image_fp, "%d\n", B_max_gray_val );

	/* Write B image buffer to file, abort on error */
	if( fwrite(B_image_buffer, 1, B_image_size, B_image_fp) != B_image_size )
	  {
	    sprintf( message, "Error writing image to file:\n%s",
		     B_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    return( FALSE );
	  }

	/* Free image buffers and close image files */
	sprintf( message, "Closing image files:\n%s\n%s",
		 A_image_file, B_image_file );
	Show_Message( message, "black" );

	free( A_image_buffer );
	free( B_image_buffer );
	fclose( A_image_fp );
	fclose( B_image_fp );

      } /* End of case NOAA : */

      break;

    case METEOR: /* Decode METEOR APT images */
      {
        /* Meteor image file name */
	char Met_image_file[MAX_FILE_NAME+20];

	int
	  line_cnt,     /* Number of APT image lines processed       */
	  carrier_val,  /* Instantaneous value of 2.4 Khz subcarrier */
	  pixel_val,    /* Averaged pixel value to enter into pixbuf */
	  max_gray_val, /* Maximum gray value of Meteor APT image    */
	  Met_image_height, /* Height in lines of Meteor APT image   */
	  Met_image_size;   /* Current size in bytes of Meteor image */
	  
	/* Buffer for the output image */
	unsigned char *Met_image_buffer;

	/* Meteor image file pointer */
	FILE *Met_image_fp;

	/* Process file name */
	if( strlen(gbl_image_file) == 0 )
	  strcpy( gbl_image_file, gbl_samples_file );
	else
	  File_Name( gbl_samples_file );

	/* Prepare image file name */
	strcpy( Met_image_file, gbl_image_file );
	strcat( Met_image_file, "-Meteor.pgm" );

	/* Print some information */
	Show_Message( "Decoding image:", "black" );
	sprintf( message, "%s", Met_image_file );
	Show_Message( message, "black" );

	/* Clear variables before decoding sample buffer */
	Met_image_height = 0;
	max_gray_val     = 0;
	Met_image_size   = 0;
	Met_image_buffer = NULL;
	carrier_val      = 0;
	line_cnt         = 0;

	/* Process samples from file to make an image */
	while( File_Fill_Buffer() != -1 )
	  {
	    /* Bypass sync and space/markers of image */
	    gbl_buffer_idx += METEOR_SYNC_IMAGE_SPACE;

	    /* Re-allocate memory to image buffer, abort on error */
	    Met_image_buffer = realloc( Met_image_buffer,
					Met_image_size + METEOR_IMAGE_WIDTH);

	    if( Met_image_buffer == NULL )
	      {
		Show_Message( "A memory allocation request has failed - Quit!",
			      "red" );
		Error_Dialog();
		return( FALSE );
	      }

	    /* Decode an image line */
	    Meteor_Image_Line( Met_image_buffer + Met_image_size, &max_gray_val );
	    Met_image_height++;
	    Met_image_size = METEOR_IMAGE_WIDTH * Met_image_height;

	    line_cnt++;

	  } /* End of while( File_Fill_Buffer() != -1 ) */

	/* Carry out histogram normalization if not suppressed */
	if( isFlagSet(IMAGE_NORMALIZE) )
	  {
	    Normalize( Met_image_buffer, Met_image_size );
	    max_gray_val = 255;
	  }

	/* Rotate images 180 degrees if requested */
	if( isFlagSet(IMAGE_ROTATE) )
	  Rotate( Met_image_buffer, Met_image_size );

	/*** Processing finished, complete image files ***/

	/* Resize window as required */
	gbl_duration = line_cnt / 2;
	Initialize_Top_window();

	/* Draw images */
	for( idy = 0; idy < gbl_duration; idy++ )
	  for( idx = 0; idx < METEOR_IMAGE_WIDTH/2; idx++ )
	    {
	      pixel_val  = 0;
 	      pixel_val += Met_image_buffer[(2*idx)   +
					    (2*idy)   * METEOR_IMAGE_WIDTH];
 	      pixel_val += Met_image_buffer[(2*idx+1) +
					    (2*idy)   * METEOR_IMAGE_WIDTH];
 	      pixel_val += Met_image_buffer[(2*idx)   +
					    (2*idy+1) * METEOR_IMAGE_WIDTH];
 	      pixel_val += Met_image_buffer[(2*idx+1) +
					    (2*idy+1) * METEOR_IMAGE_WIDTH];

 	      pixel_val /= 4;

	      /* Find the A pixel at x,y */
	      pixel = gbl_pixel_buf +
		     (idy*gbl_rowstride) +
		     (idx*gbl_n_channels);

	      /* Assign image R,G,B,A to grayscale value */
	      pixel[0] = pixel_val;
	      pixel[1] = pixel_val;
	      pixel[2] = pixel_val;
	      pixel[3] = 0xff;

	    } /* for( idx = 0; idx < line_cnt; idx++ ) */

	/* Draw finished image */
	gtk_widget_queue_draw( gbl_apt_image );

	/* Open Meteor APT image file */
	Met_image_fp  = fopen( Met_image_file, "w+" );
	if( Met_image_fp == NULL )
	  {
	    perror( Met_image_file );
	    sprintf( message, "Failed to open file for writing:\n%s",
		     Met_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    return( FALSE );   
	  }

	/* Write header in output Meteor PGM files */
	fprintf( Met_image_fp, "P5\n" );
	fprintf( Met_image_fp, "%d %d\n",
		 METEOR_IMAGE_WIDTH, Met_image_height );
	fprintf( Met_image_fp, "%d\n", max_gray_val );

	/* Write image buffer to file, abort on error */ 
	if( fwrite(Met_image_buffer, 1, Met_image_size, Met_image_fp) !=
	    Met_image_size )
	  {
	    sprintf( message, "Error writing image to file:\n%s",
		     Met_image_file );
	    Show_Message( message, "red" );
	    Error_Dialog();
	    return( FALSE );
	  }

	/* Free image buffer and close image file */
	sprintf( message, "Closing image file:\n%s", Met_image_file );
	Show_Message( message, "black" );

	free( Met_image_buffer );
	fclose( Met_image_fp );

      } /* End of case METEOR : */

    } /* End of switch( gbl_sat_type ) */

  Cleanup();

  return( FALSE );

} /* End of File_Process_Image() */

/*------------------------------------------------------------------------*/

/*  Normalize()
 *
 *  Does histogram (linear) normalization of a pgm (P5) image file
 */

void
Normalize( unsigned char *image_buffer, int image_size )
{
  int
    hist[256],  /* Intensity histogram of pgm image file  */
    blk_cutoff, /* Count of pixels for black cutoff value */
    wht_cutoff, /* Count of pixels for white cutoff value */
    pixel_val,  /* Used for calculating normalized pixels */
    pixel_cnt,  /* Total pixels counter for cut-off point */
    idx;        /* Index for loops etc */

  unsigned char
    black_val,  /* Black cut-off pixel intensity value */
    white_val,  /* White cut-off pixel intensity value */
    val_range;  /* Range of intensity values in image  */

  if( image_size <= 0 )
    {
      Show_Message( "Image file seems empty! - "
		    "Normalization not performed.", "red" );
      return;
    }

  /* Clear histogram */
  for( idx = 0; idx < 256; idx++ )
    hist[ idx ] = 0;

  /* Build image intensity histogram */
  for( idx = 0; idx < image_size; idx++ )
    ++hist[ image_buffer[idx] ];

  /* Determine black/white cut-off counts */
  blk_cutoff = (image_size * BLACK_CUT_OFF) / 100;
  wht_cutoff = (image_size * WHITE_CUT_OFF) / 100;

  /* Find black cut-off intensity value */
  pixel_cnt = 0;
  for( black_val = 0; black_val < 256; black_val++ )
    {
      pixel_cnt += hist[ black_val ];
      if( pixel_cnt > blk_cutoff )
	break;
    }

  /* Find white cut-off intensity value */
  pixel_cnt = 0;
  for( white_val = 255; white_val >= 1; white_val-- )
    {
      pixel_cnt += hist[ white_val ];
      if( pixel_cnt > wht_cutoff )
	break;
    }

  /* Rescale pixels in image for full intensity range */
  val_range = white_val - black_val;
  if( (int) val_range <= 0 )
    {
      Show_Message( "Image seems flat! - "
		    "Normalization not performed.", "red" );
      return;
    }

  /* Perform histogram normalization on images */
  Show_Message( "Performing histogram normalization ... ", "black" );

  for( pixel_cnt = 0; pixel_cnt < image_size; pixel_cnt++ )
    { 
      pixel_val = image_buffer[ pixel_cnt ];
      pixel_val = ( (pixel_val - black_val) * 255 ) / val_range;

      pixel_val = ( pixel_val < 0 ? 0 : pixel_val );
      pixel_val = ( pixel_val > 255 ? 255 : pixel_val );
      image_buffer[ pixel_cnt ] = pixel_val;
    }

} /* End of Normalize() */

/*------------------------------------------------------------------------*/

/*  Rotate()
 *
 *  Rotates a pgm (P5) image by 180 degrees
 */

void
Rotate( unsigned char *image_buffer, int image_size )
{
  int idx; /* Index for loops etc */

  unsigned char
    *idx_temp,  /* Buffer location to be saved in temp    */
    *idx_swap;  /* Buffer location to be swaped with temp */

  /* Holds a pixel value temporarily */
  unsigned char temp;

  if( image_size <= 0 )
    {
      Show_Message( "Image file empty! - Rotation not performed", "red");
      return;
    }

  /* Rotate image 180 degrees */
  Show_Message( "Rotating image by 180 degrees ... ", "black" );

  for( idx = 0; idx < image_size / 2; idx++ )
    {
      idx_temp = image_buffer + idx;
      idx_swap = image_buffer - 1 + image_size - idx;
      temp = *( idx_temp );
      *( idx_temp ) = *( idx_swap );
      *( idx_swap ) = temp;
    }

} /* End of Rotate() */

/*------------------------------------------------------------------------*/

/*  Scalar_Mag()
 *
 *  Since sampling of the 2.4 Khz sub-carrier is done at 9.6 KHz
 *  consecutive samples are at 90 deg. The instantaneous magnitude
 *  of the carrier is sqrt(x^2+y^2) were x and y are two 
 *  consecutive samples. This is approximated by a Taylor 
 *  series expansion in (x+y) to avoid floating point sqrt,
 *  e.g: ((x+y)^2-x*y)/(x+y) to first order, and refined by
 *  one iteration of the Newton-Raphson method. This is rather
 *  wierd but seems to give the most accurate approximation.
 */

unsigned char
Scalar_Mag( unsigned char x, unsigned char y )
{
  /* Scalar magnitude of two samples at 90 deg */
  unsigned char scalar_mag;
  
  int
    sum,    /* Sum of two consecutive samples    */
    prod,   /* Product of two cnsecutive samples */
    aprox;  /* First approxim. to sqrt(x^2+y^2)  */

  /* Remove the 'neutral' value (0x80)   */
  /* of the unsigned 8-bit audio samples */
  x = abs(x - 128);
  y = abs(y - 128);

  /* Calculate first approximation to sqrt(x^2+y^2) */
  sum   = x + y;
  prod  = x * y;
  aprox = ( sum == 0 ? 0 : (sum*sum - prod)/sum );

  /* Refine by one iteration using Newton-Raphson method */
  scalar_mag = ( aprox == 0 ? 0 :
	       ((aprox*aprox+sum*sum-(prod*2))/aprox)/2 );

  return( scalar_mag );

} /* End of Scalar_Mag() */

/*------------------------------------------------------------------------*/

/*  Set_Sync_Icon()
 *
 * Changes the sync status icon
 */

void
Set_Sync_Icon( gchar *icon_name )
{
  /* Remove previous icon */
  gtk_container_remove( GTK_CONTAINER(gbl_status_table), gbl_sync_icon );

  /* Attach a new icon */
  gbl_sync_icon = gtk_image_new_from_stock( icon_name, GTK_ICON_SIZE_BUTTON );
  gtk_widget_show( gbl_sync_icon );

  gtk_table_attach( GTK_TABLE(gbl_status_table), gbl_sync_icon, 0, 1, 0, 1,
                    (GtkAttachOptions) (GTK_FILL),
                    (GtkAttachOptions) (GTK_FILL), 0, 0 );

} /* Set_Sync_Icon( gchar *img_name ) */

/*------------------------------------------------------------------------*/
