/*  main.c
 *  
 *  Main body of xwxapt application
 */  

/*
 *  xwxapt: An application to decode APT signals from
 *  weather satellites and produce image(s) of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or 
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>

#include "interface.h"
#include "support.h"
#include "xwxapt.h"


GtkWidget
  *xwxapt_top_window, /* xwxapt's top window   */
  *gbl_text_scroller, /* Text view scroller    */
  *gbl_level_scale,   /* Audio level scale     */
  *gbl_sync_scale,    /* Sync level scale      */
  *gbl_status_table,  /* Status widgets' table */
  *gbl_sync_icon,     /* Sync status indicator */
  *gbl_apt_image,     /* Image to be displayed */
  *gbl_sync_label;    /* Sync status ind label */

/* Text buffer for text view */
GtkTextBuffer *gbl_text_buffer;

/* Pixel buffer for display */
GdkPixbuf *gbl_pixbuf = NULL;

/* Pointer to pixbuf pixels */
guchar *gbl_pixel_buf;

/* Pixbuf rowstride and num of channels */
gint
  gbl_rowstride,
  gbl_n_channels;

/* Satellite type names */
char *gbl_sat_names[] = { SAT_NAMES };

/* List of required sync position references */
int gbl_sync_refs[] = { SYNC_REFS };

int
  gbl_sat_type,   /* Weather satellite type number   */
  gbl_duration,   /* Time limit in sec on processing */
  gbl_buffer_idx, /* Index to audio samples buffer   */
  gbl_sync_ref,   /* Required position of sync train */
  gbl_dsp_fd=0,   /* File descriptor of DSP device   */
  gbl_mix_fd=0;   /* File descriptor of mixer device */

char
  gbl_samples_file[MAX_FILE_NAME+1], /* Samples file name */
  gbl_image_file  [MAX_FILE_NAME+1]; /* Image files name  */

/* Pointer to samples file */
FILE *gbl_samples_fp = NULL;

/* Signal handler */
static void sig_handler( int signal );


int
main (int argc, char *argv[])
{
  /* Adjustment for scales/scrollers */
  GtkAdjustment *adjustment;

  /* Command line option returned by getopt() */ 
  int option;       

  /* getopt() variables */
  extern char *optarg;
  extern int optind, opterr, optopt;

  /* New and old actions for sigaction() */
  struct sigaction sa_new, sa_old;

  /* Initialize new actions */
  sa_new.sa_handler = sig_handler;
  sigemptyset( &sa_new.sa_mask );
  sa_new.sa_flags = 0;

  /* Register function to handle signals */
  sigaction( SIGINT,  &sa_new, &sa_old );
  sigaction( SIGSEGV, &sa_new, 0 );
  sigaction( SIGFPE,  &sa_new, 0 );
  sigaction( SIGTERM, &sa_new, 0 );
  sigaction( SIGABRT, &sa_new, 0 );

  /* Start GTK+ */
  gtk_set_locale();
  gtk_init( &argc, &argv );

  /* Process command line options */
  while( (option = getopt(argc, argv, "hv") ) != -1 )
    switch( option )
      {
      case 'h' : /* Print usage and exit */
	Usage();
	return(0);

      case 'v' : /* Print version */
	fprintf( stderr, "%s version %s\n", PACKAGE, VERSION );
	return(0);

      default: /* Print usage and exit */
	Usage();
	exit(-1);

      } /* End of switch( option ) */

  /* Defaults/initialization */
  gbl_duration        = DEFAULT_DURATION;
  gbl_sat_type        = NOAA;
  gbl_image_file[0]   = '\0';
  gbl_samples_file[0] = '\0';
  SetFlag( IMAGE_NORMALIZE );

  /* Create xwxapt window */
  xwxapt_top_window = create_top_window();

  /* Create global sync status widgets' table */
  gbl_status_table = lookup_widget( xwxapt_top_window, "table1" );

  /* Create global status indicator label */
  gbl_sync_label = lookup_widget( xwxapt_top_window, "label5" );

  /* Create global sync status indicator icon */
  gbl_sync_icon = lookup_widget( xwxapt_top_window, "image2" );

  /* Create the text view scroller */
  gbl_text_scroller = lookup_widget( xwxapt_top_window, "scrolledwindow2" );

  /* Get text buffer */
  gbl_text_buffer = gtk_text_view_get_buffer
    ( GTK_TEXT_VIEW(lookup_widget(xwxapt_top_window, "textview1")) );

  /* Create some rendering tags */
  gtk_text_buffer_create_tag( gbl_text_buffer, "black",
			      "foreground", "black", NULL);  
  gtk_text_buffer_create_tag( gbl_text_buffer, "red",
			      "foreground", "red", NULL);  
  gtk_text_buffer_create_tag( gbl_text_buffer, "green",
			      "foreground", "darkgreen", NULL);  
  gtk_text_buffer_create_tag( gbl_text_buffer, "bold",
			      "weight", PANGO_WEIGHT_BOLD, NULL);  

  /* Create the sliding scales */
  gbl_level_scale = lookup_widget( xwxapt_top_window, "hscale1" );
  adjustment = gtk_range_get_adjustment( GTK_RANGE(gbl_level_scale) );
  adjustment->lower = -20.0;
  adjustment->upper = +20.0;

  gbl_sync_scale  = lookup_widget( xwxapt_top_window, "hscale2" );
  adjustment = gtk_range_get_adjustment( GTK_RANGE(gbl_sync_scale) );
  adjustment->lower = 0.0;
  adjustment->upper = 2500.0;

  /* Initialize top window etc */
  Initialize_Top_window();

  /* Open xwxapt window */
  gtk_widget_show( xwxapt_top_window );
  Show_Message( "Welcome to xwxapt version "VERSION, "bold" );

  gtk_main ();

  return 0;
}

/*------------------------------------------------------------------------*/

/*  Initialize_Top_window()
 *
 *  Initializes xwxapt's top window
 */

void
Initialize_Top_window( void )
{
  /* The scrolled window image container */
  GtkWidget *image_scroller;

  if( gbl_pixbuf != NULL )
    g_object_unref( gbl_pixbuf );

  switch( gbl_sat_type )
    {
    case NOAA: /* Create pixbuff for NOAA images */
      gbl_pixbuf = gdk_pixbuf_new( GDK_COLORSPACE_RGB, FALSE, 8,
				   NOAA_IMAGE_WIDTH, gbl_duration );
      break;

    case METEOR: /* Create pixbuff for METEOR images */
      gbl_pixbuf = gdk_pixbuf_new( GDK_COLORSPACE_RGB, FALSE, 8,
				   METEOR_IMAGE_WIDTH/2, gbl_duration );
    }

  /* Error, not enough memory */
  if (gbl_pixbuf == NULL)
    {
      Show_Message( "Failed to allocated memory to pixbuf - Quit!", "red" );
      Error_Dialog();
      return;
    }

  /* Fill pixbuf with background color */
  gdk_pixbuf_fill( gbl_pixbuf, 0xaaaaaaff );

  /* Get details of pixbuf */
  gbl_pixel_buf  = gdk_pixbuf_get_pixels(     gbl_pixbuf );
  gbl_rowstride  = gdk_pixbuf_get_rowstride(  gbl_pixbuf );
  gbl_n_channels = gdk_pixbuf_get_n_channels( gbl_pixbuf );

  /* Globalize image to be displayed */
  gbl_apt_image = lookup_widget( xwxapt_top_window, "image1" );

  /* Set apt image from pixbuff */
  gtk_image_set_from_pixbuf( GTK_IMAGE(gbl_apt_image), gbl_pixbuf );
  gtk_widget_show( gbl_apt_image );

  /* Set window size as required */
  image_scroller = lookup_widget( xwxapt_top_window, "scrolledwindow1" );
  gtk_widget_set_size_request( image_scroller,
			       (gbl_duration <= DEFAULT_DURATION ?
			        NOAA_IMAGE_WIDTH + 8 : NOAA_IMAGE_WIDTH + 26),
			       (gbl_duration <= DEFAULT_DURATION ?
				gbl_duration+8 : DEFAULT_DURATION+8) );

  gtk_window_resize( GTK_WINDOW(xwxapt_top_window), 10, 10 );

} /* Initialize_Top_window() */

/*------------------------------------------------------------------------*/

/*  File_Name()
 *
 *  Prepare a file name from optarg, use date and time if null
 */

void
File_Name( char *file_name )
{
  int
    len, /* String length of file_name */
    idx; /* Index to while loop */

  /* If file_name is null, use date and time as file name */
  if( strlen(file_name) == 0 )
    {
      /* Variables for reading time (UTC) */
      time_t tp;
      struct tm utc;

      /* Prepare a file name as UTC date-time. */
      /* Default paths are images/ and record/ */
      time( &tp );
      utc = *gmtime( &tp );

      if( isFlagSet(ACTION_PROCESS_DSP) )
	strftime( file_name, 24, "images/%d%b%Y-%H%M", &utc );
      else
	if( isFlagSet(ACTION_RECORD_APT) )
	  strftime( file_name, 24, "record/%d%b%Y-%H%M", &utc );
    }
  else /* Remove leading spaces from file_name */
    {
      idx = 0;
      len = strlen(file_name);
      do idx++;
      while( (file_name[idx] == ' ') && (idx != len) );
      file_name += idx - 1;
    }

} /* End of File_Name() */

/*------------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

void
Usage( void )
{
  fprintf( stderr, "%s\n", "Usage: xwxapt -hv" );

  fprintf( stderr, "%s\n\n",
  "       -h: Print this usage information and exit.");

  fprintf( stderr, "%s\n\n",
  "       -v: Print version number and exit.");

} /* End of Usage() */

/*------------------------------------------------------------------------*/

/*  Show_Message()
 *
 *  Prints a message string in the Text View scroller
 */

void
Show_Message( char *message, char *attr )
{
  GtkAdjustment *adjustment;

  static GtkTextIter iter;
  static gboolean first_call = TRUE;

  /* Initialize */
  if( first_call )
    {
      first_call = FALSE;
      gtk_text_buffer_get_iter_at_offset( gbl_text_buffer, &iter, 0 );
    }

  /* Print message */
  gtk_text_buffer_insert( gbl_text_buffer, &iter, "\n", -1 );

  gtk_text_buffer_insert_with_tags_by_name( gbl_text_buffer, &iter, 
					    message, -1, attr, NULL );

  /* Scroll Text View to bottom */
  adjustment = gtk_scrolled_window_get_vadjustment
    ( GTK_SCROLLED_WINDOW(gbl_text_scroller) );

  gtk_adjustment_set_value( adjustment, adjustment->upper );

  gtk_widget_grab_focus(gbl_text_scroller);

} /* End of Show_Message() */

/*------------------------------------------------------------------------*/

/*  Cleanup()
 *
 *  Cleanup before quitting or not using sound card
 */

void
Cleanup( void )
{
  if( gbl_mix_fd != 0 )
    {
      close( gbl_mix_fd );
      gbl_mix_fd = 0;
    }

  if( gbl_dsp_fd != 0 )
    {
      close( gbl_dsp_fd );
      gbl_dsp_fd = 0;
    }

  if( gbl_samples_fp != NULL )
    {
      fclose( gbl_samples_fp );
      gbl_samples_fp = NULL;
    }

  gbl_image_file[0]   = '\0';
  gbl_samples_file[0] = '\0';

  /* Clear status indications */
  Set_Sync_Icon( "gtk-stop" );
  gtk_label_set_label( GTK_LABEL(gbl_sync_label), "Stop" );
  gtk_range_set_value( GTK_RANGE(gbl_level_scale), 0.0 );
  gtk_range_set_value( GTK_RANGE(gbl_sync_scale),  0.0 );

  ClearFlag( SOUND_CARD_SETUP );
  ClearFlag( ACTION_FLAGS_ALL );
  ClearFlag( ICON_FLAGS_ALL );

} /*  Cleanup() */

/*------------------------------------------------------------------------*/

/*  sig_handler()
 *
 *  Signal Action Handler function
 */

static void sig_handler( int signal )
{
  Cleanup();

  switch( signal )
    {
    case SIGINT :
      fprintf( stderr, "\n%s\n", "xwxapt: Exiting via User Interrupt" );
      exit(-1);
      break;

    case SIGSEGV :
      fprintf( stderr, "\n%s\n", "xwxapt: Segmentation Fault" );
      exit(-1);
      break;

    case SIGFPE :
      fprintf( stderr, "\n%s\n", "xwxapt: Floating Point Exception" );
      exit(-1);
      break;

    case SIGABRT :
      fprintf( stderr, "\n%s\n", "xwxapt: Abort Signal received" );
      exit(-1);
      break;

    case SIGTERM : 
      fprintf( stderr, "\n%s\n", "xwxapt: Termination Request received" );
      exit(-1);
    }

} /* End of sig_handler() */

/*------------------------------------------------------------------------*/

/* Functions for testing and setting/clearing flags */

/* An int variable holding the single-bit flags */
static int Flags = 0;

int
isFlagSet(int flag)
{
  return (Flags & flag);
}

int
isFlagClear(int flag)
{
  return (~Flags & flag);
}

void
SetFlag(int flag)
{
  Flags |= flag;
}

void
ClearFlag(int flag)
{
  Flags &= ~flag;
}

void
ToggleFlag(int flag)
{
  Flags ^= flag;
}

/*------------------------------------------------------------------*/
