/*  sound.c
 *  
 *  Sound-card handling functions of xwxapt application
 */  

/*
 *  xwxapt: An application to decode APT signals from weather 
 *  satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or 
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include <gtk/gtk.h>
#include "xwxapt.h"
#include "interface.h"
#include "support.h"

extern GtkWidget
  *xwxapt_top_window, /* xwxapt top window  */
  *gbl_text_scroller, /* Text view scroller */
  *gbl_level_scale,   /* Audio level scale  */
  *gbl_sync_scale,    /* Sync level scale   */
  *gbl_sync_label,    /* Sync status label  */
  *gbl_apt_image;     /* Decoded Wx image   */

/* Text buffer for text view */
extern GtkTextBuffer *gbl_text_buffer;

/* Satellite type names */
extern char *gbl_sat_names[];

/* List of required sync position references */
extern int gbl_sync_refs[];

extern int
  gbl_sat_type,   /* Weather satellite type number   */
  gbl_duration,   /* Time limit in sec on processing */
  gbl_buffer_idx, /* Index to audio samples buffer   */
  gbl_sync_ref,   /* Required position of sync train */
  gbl_dsp_fd,     /* File descriptor of DSP device   */
  gbl_mix_fd;     /* File descriptor of mixer device */

/* Pointer to samples file */
extern FILE *gbl_samples_fp;

/* Audio/DSP samples buffer for one line */
extern unsigned char gbl_line_buffer[];

/* Buffer for reading in DSP data */
static unsigned char *dsp_buffer;

static int
  dsp_buf_idx, /* Index for this buffer  */
  chunk_size;  /* DSP block (chunk) size */

char
  gbl_samples_file[MAX_FILE_NAME+1],  /* Sample file name */
  gbl_image_file  [MAX_FILE_NAME+1];  /* Image file name  */

/* dsp sampling rate error */
int dsp_rate_err;

/*------------------------------------------------------------------------*/

/*  Setup_Sound_Card()
 *
 *  Sets up mixer and DSP device and adjusts input
 *  level for correct sampling of APT signal
 */

void
Setup_Sound_Card( void )
{
  int
    rec_src,     /* Recording/Capture source (Line?)  */
    reclev_dev,  /* Recording Level device (IGAIN?)   */
    record_lev,  /* Recording level request to mixer  */
    dsp_speed,   /* DSP sampling speed (9600 Hz)      */
    sample_fmt,  /* DSP sample format, 8-bit unsigned */
    channels,    /* (Stereo) or Mono mode selection   */
    itmp;        /* Temporary int for tests etc       */

  /* Sound device names from souncard.h */
  char *dev_name[SOUND_MIXER_NRDEVICES] = SOUND_DEVICE_NAMES;

  /* Messages buffer */
  char message[350];

  /* File lock information */
  struct flock lockinfo;


  /* Return if sound card is ready */
  if( isFlagSet(SOUND_CARD_SETUP) )
    return;

  /*** Open and lock mixer and dsp devices ***/

  Show_Message( "Setting up the sound card ... ", "black" );

  /* Open mixer device, abort on error */
  Show_Message( "Opening mixer device: "MIXER_DEV, "black" );
  if( (gbl_mix_fd = open(MIXER_DEV, O_RDWR, 0)) == -1 )
    {
      perror( "xwxapt: "MIXER_DEV );
      Show_Message( "Failed to open "MIXER_DEV, "red" );
      Error_Dialog();
      return;
    }

  /* Attempt to lock entire mixer device file */
  Show_Message( "Locking mixer device: "MIXER_DEV, "black" );
  lockinfo.l_type   = F_WRLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If mixer device is already locked, abort */
  if( fcntl( gbl_mix_fd, F_SETLK, &lockinfo ) < 0 )
    {
      perror( "xwxapt: Lock "MIXER_DEV );
      fcntl( gbl_mix_fd, F_GETLK, &lockinfo );
      sprintf( message,
	       "Failed to lock "MIXER_DEV": Device already locked by pid %d.",
	       lockinfo.l_pid);
      Show_Message( message, "red" );
      Error_Dialog();
      return;
    }

  /* Open DSP device, abort on error */
  Show_Message( "Opening dsp device: "DSP_DEV, "black" );
  if( (gbl_dsp_fd = open( DSP_DEV, O_RDONLY, 0)) == -1 )
    {
      perror( "xwxapt: "DSP_DEV );
      Show_Message( "Failed to open "DSP_DEV, "red" );
      Error_Dialog();
      return;
    }

  /* Attempt to lock entire DSP device file */
  Show_Message( "Locking dsp device: "DSP_DEV, "black" );
  lockinfo.l_type   = F_RDLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If DSP device is already locked, abort */
  if( fcntl( gbl_dsp_fd, F_SETLK, &lockinfo ) < 0 )
    {
      perror( "xwxapt: Lock "DSP_DEV );
      fcntl( gbl_dsp_fd, F_GETLK, &lockinfo );
      sprintf( message, "Lock "DSP_DEV": Device already locked by pid %d",
	       lockinfo.l_pid);
      Show_Message( message, "red" );
      Error_Dialog();
      return;
    }

  /*** Set up sound card parameters ***/

  Show_Message( "Setting up sound card parameters...", "black" );

  /* Find number of recording source matching name in xwxapt.h */
  for( rec_src = 0; rec_src < SOUND_MIXER_NRDEVICES; rec_src++ )
    if( strcmp( dev_name[rec_src], RECORD_SRC ) == 0 )
      break;

  if( rec_src == SOUND_MIXER_NRDEVICES )
    {
      Show_Message( "Failed to find device: "RECORD_SRC, "red" );
      Error_Dialog();
      return;
    }

  /* Make recording source mask */
  itmp = 1 << rec_src;

  /* Attempt to select recording source, abort on error */
  if( ioctl(gbl_mix_fd, SOUND_MIXER_WRITE_RECSRC, &itmp) == -1 ) 
    {
      perror("xwxapt: SOUND_MIXER_WRITE_RECSRC");
      Show_Message( "Failed to select device: "RECORD_SRC, "red" );
      Error_Dialog();
      return;
    }
  if( itmp != 1 << rec_src )
    {
      Show_Message( "Failed to select device: "RECORD_SRC, "red" );
      Error_Dialog();
      return;
    }

  /* Write default recording level to recording source, abort on error */
  record_lev = RECORD_LEV;
  if( ioctl(gbl_mix_fd, MIXER_WRITE(rec_src), &record_lev) == -1 )
    {
      perror("xwxapt: Setting Recording Source level");
      Show_Message( "Failed to set Recording Source level.", "red" );
      Error_Dialog();
      return;
    }
  if( record_lev != RECORD_LEV )
    {
      Show_Message( "Failed to set Recording Source level.", "red" );
      Error_Dialog();
      return;
    }

  /* Find number of recording level device matching name in xwxapt.h */
  for( reclev_dev = 0; reclev_dev < SOUND_MIXER_NRDEVICES; reclev_dev++ )
    if( strcmp( dev_name[reclev_dev], RECLEV_DEV ) == 0 )
      break;

  if( reclev_dev ==  SOUND_MIXER_NRDEVICES )
    {
      Show_Message( "Failed to select device: "RECLEV_DEV, "red" );
      Error_Dialog();
      return;
    }

  /* Write default level to recording level device, abort on error */
  record_lev = RECORD_LEV;
  if( ioctl(gbl_mix_fd, MIXER_WRITE(reclev_dev), &record_lev) == -1 )
    {
      perror("xwxapt: Setting Recording level");
      Show_Message( "Failed to set recording level.", "red" );
      Error_Dialog();
      return;
    }
  if( record_lev != RECORD_LEV )
    {
      Show_Message( "Failed to set Recording level.", "red" );
      Error_Dialog();
      return;
    }

  /* Setup dsp fragment size to 4k, ignore errors */
  chunk_size = (NUM_OF_CHANNELS << 16) + 12;
  ioctl( gbl_dsp_fd, SNDCTL_DSP_SETFRAGMENT, &chunk_size );

  /* Set sample format to unsigned 8-bit, abort on error */
  sample_fmt = AFMT_U8;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_SETFMT, &sample_fmt) == -1 )
    {
      perror( "xwxapt: SNDCTL_DSP_SETFMT");
      Show_Message( "Failed to set the dsp sample format.", "red" );
      Error_Dialog();
      return;
    }
  if( sample_fmt != AFMT_U8 )
    {
      Show_Message( "Failed to set the dsp sample format.", "red" );
      Error_Dialog();
      return;
    }

  /* Set stereo or mono mode */
  channels = NUM_OF_CHANNELS;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_CHANNELS, &channels) == -1 )
    {
      perror("xwxapt: SNDCTL_DSP_CHANNELS");
      Show_Message( "Failed to set the dsp number of channels.", "red" );
      Error_Dialog();
      return;
    }
  if( channels != NUM_OF_CHANNELS )
    {
      Show_Message( "Failed to set the dsp number of channels.", "red" );
      Error_Dialog();
      return;
    }

  /* Setup DSP sampling speed, abort on error */
  dsp_speed = DSP_SPEED;
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_SPEED, &dsp_speed) == -1 )
    {
      perror( "xwxapt: SNDCTL_DSP_SPEED");
      Show_Message( "Failed to set the dsp sampling speed.", "red" );
      Error_Dialog();
      return;
    }
  if( dsp_speed != DSP_SPEED )
    {
      Show_Message( "Failed to set the dsp sampling speed.", "red" );
      Error_Dialog();
      return;
    }

  /* Read DSP block (chunck) size */
  if( ioctl(gbl_dsp_fd, SNDCTL_DSP_GETBLKSIZE, &chunk_size ) == -1 )
    {
      perror("xwxapt: Reading dsp block (chunk) size");
      Show_Message( "Failed to read dsp block (chunk) size.", "red" );
      Error_Dialog();
      return;
    }

  /* Allocate memory to DSP data buffer */
  dsp_buffer = malloc( chunk_size );
  if( dsp_buffer == NULL )
    {
      Show_Message( "Failed to malloc memory to samples buffer - Quit!", "red" );
      Error_Dialog();
      return;
    }

  /* Succeeded, show a message and set flag */
  Show_Message( "Sound card set-up completed successfully!", "green");
  SetFlag( SOUND_CARD_SETUP );

  return;

} /* End of Setup_Sound_Card() */

/*------------------------------------------------------------------------*/

/*  Setup_Audio_Level()
 *
 *  Displays a simple graphical aid for setting
 *  up the audio level from the APT receiver
 */

gboolean
Setup_Audio_Level( gpointer data )
{
  int
    carrier_max, /* Max value of carrier in  buffer */
    carrier_val, /* Sampled value of 2.4kHz carrier */
    level_error, /* Error in the input audio level  */
    idx;         /* Index used in loops etc         */

  /* First idle call */
  static gboolean first_call = TRUE;

  /* Setup sound card, abort on error */
  if( first_call )
    {
      Setup_Sound_Card();
      if( isFlagClear(SOUND_CARD_SETUP) )
	return( FALSE );

      Show_Message( "Starting the Audio Level Setup function.", "black" );
      first_call = FALSE;
    }

  /* Fill in the samples buffer from DSP, abort on error */
  if( Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE) == -1 )
    {
      Show_Message( "Failed to fill samples buffer from dsp.", "red" );
      Show_Message( "Stopping Audio Level Setup function.", "black" );
      Error_Dialog();
      first_call = TRUE;
      return( FALSE );
    }

  /*** Find max value of 2.4kHz subcarrier in buffer ***/
  carrier_max = 0;

  for( idx = 0; idx < BLOCK_SIZE-1; idx++ )
    {
      carrier_val = Scalar_Mag(gbl_line_buffer[idx],
			       gbl_line_buffer[idx+1]);

      /* Find maximum carrier level */
      if( carrier_val > carrier_max )
	carrier_max = carrier_val;
    }

  /* Find input audio level error and adjust scale */
  level_error = 0;
  switch( gbl_sat_type )
    {
    case NOAA:
      level_error = carrier_max - NOAA_CARRIER_MAX;
      break;

    case METEOR:
      level_error = carrier_max - METEOR_CARRIER_MAX;
    }

  gtk_range_set_value( GTK_RANGE(gbl_level_scale), (gdouble)level_error );

  if( isFlagSet(ACTION_SETUP_AUDIO) )
    return( TRUE );

  Show_Message( "Stopping Audio Level Setup function.", "black" );
  Cleanup();
  first_call = TRUE;

  return( FALSE );

} /* Setup_Audio_Level() */

/*------------------------------------------------------------------------*/

/*  Synchronize_Buffer()
 *
 *  Synchronizes samples buffer with the APT sync train
 */

gboolean
Synchronize_Buffer( gpointer data )
{
  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Messages buffer */
  char message[350];

  int
    block_size; /* Data block size used as read() argument */


  static int
    sync_err = 0,     /* Sync position error relative to refer */
    ave_sync_err = 5; /* Average of sync train position error  */


  /* Cancel on user request */
  if( isFlagClear(ACTION_SYNC_BUFFER) )
    {
	Show_Message( "Search for APT sync train stopped.", "black" );
	Cleanup();
	first_call = TRUE;
	return( FALSE );
      }

  if( first_call )
    {
      /* Setup sound card, abort on error */
      Setup_Sound_Card();
      if( isFlagClear(SOUND_CARD_SETUP) )
	return( FALSE );

      /* Print some information */
      Show_Message( "Synchronizing samples buffer"
		    " with APT sync train ...", "black" );

      sprintf( message, "Looking for a %s type sync train ... ",
	       gbl_sat_names[gbl_sat_type] );
      Show_Message( message, "black" );

      /* Load appropriate sync reference */
      gbl_sync_ref = gbl_sync_refs[ gbl_sat_type ];

      /* Fill in the line buffer from DSP */
      if( Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE) == -1 )
	{
	  Show_Message( "Failed to fill line buffer from dsp.", "red" );
	  Show_Message( "Stopping Buffer Sync function.", "black" );
	  Error_Dialog();
	  first_call = TRUE;
	  return( FALSE );
	}
      sync_err = gbl_buffer_idx - gbl_sync_ref;

      /* To avoid false sync detection */
      ave_sync_err = 5;

      first_call = FALSE;

    } /* if( first_call ) */

  /* Set buffer index to just after sync train */
  gbl_buffer_idx = 0;
  Detect_Sync_Train( BLOCK_SIZE );

  /* Coarse adjustment of the sync train position in the */
  /* samples buffer is done by dummy-reading dsp samples */
  block_size = sync_err;
  if( block_size < 0 )
    block_size += BLOCK_SIZE;
  block_size += (block_size * dsp_rate_err) / BLOCK_SIZE;

  /* Fill in the samples buffer from DSP. This is a 'dummy' read */
  /* to dump most of the samples before the sync train position  */
  Read_Dsp_Buffer(gbl_line_buffer, block_size );

  /* Fill in the samples buffer from DSP */
  if( Read_Dsp_Buffer(gbl_line_buffer, BLOCK_SIZE+dsp_rate_err) == -1 )
    {
      Show_Message( "Failed to fill line buffer from dsp.", "red" );
      Show_Message( "Stopping Buffer Sync function.", "black" );
      Error_Dialog();
      first_call = TRUE;
      return( FALSE );
    }

  /* Set buffer index to just after sync train */
  gbl_buffer_idx = 0;
  Detect_Sync_Train( BLOCK_SIZE );

  /* Error in sync train position relative to ref */
  sync_err = gbl_buffer_idx - gbl_sync_ref;

  /* If close to sync, start dsp rate error measurement */
  if( abs(sync_err) < SYNC_DETECT )
    {
      if( sync_err > 0 )
	dsp_rate_err++;
      else
	if( sync_err < 0 )
	  dsp_rate_err--;

      /* Window average of sync error */
      ave_sync_err = ( 3*ave_sync_err + 4*sync_err ) / 4;
    }

  /* Continue if sync error not zero */
  if( abs(ave_sync_err) != 0 )
    return( TRUE );

  /* Print some information */
  Show_Message( "Samples buffer synchronized"
		" with APT sync train.", "green" );

  /* Start the selected process */
  if( isFlagSet(ACTION_PROCESS_DSP) )
    gtk_idle_add( Dsp_Process_Image, NULL );
  else
    if( isFlagSet(ACTION_RECORD_APT) )
      gtk_idle_add( Record_APT_Signal, NULL );

  first_call = TRUE;
  ClearFlag(ACTION_SYNC_BUFFER);

  return( FALSE );

} /* End of Synchronize_Buffer() */

/*------------------------------------------------------------------------*/

/*  Record_APT_Signal()
 *
 *  Records samples of the 2.4 Khz subcarrier in a file
 */

gboolean
Record_APT_Signal( gpointer data )
{
  /* Number of APT image lines processed */
  static int line_cnt;

  /* First call of function flag */
  static gboolean first_call = TRUE;

  /* Messages buffer */
  char message[350];

  /* Count of samples returned by read()/write() */
  ssize_t sample_cnt;


  if( first_call )
    {
      Show_Message( "Starting recording of apt signal to file.", "black" );

      /* Add extension to file name */
      strcat( gbl_samples_file, "-" );
      strcat( gbl_samples_file, gbl_sat_names[gbl_sat_type] );
      strcat( gbl_samples_file, ".bin" );

      /* Create a file for writing samples */
      gbl_samples_fp = fopen( gbl_samples_file, "w+" );
      if( gbl_samples_fp == NULL )
	{
	  perror( gbl_samples_file );
	  sprintf( message, "Failed to open recorded samples file:\n%s",
		   gbl_samples_file );
	  Show_Message( message, "red" );
	  Error_Dialog();
	  return( FALSE );
	}

      /* Read samples from DSP, stop on user interrupt */
      /* or at time limit in number of APT lines read. */
      /* gbl_duration is in seconds of recording time. */
      sprintf( message, "Setting recording duration to %d sec.",
	       gbl_duration );
      Show_Message( message, "black" );

      /* Print some information */
      sprintf( message, "Recording APT signal in:\n%s ...",
	       gbl_samples_file );
      Show_Message( message, "black" );

      line_cnt = 0;

      /* Write satellite type to samples file, abort on error */
      if( (sample_cnt = fwrite( &gbl_sat_type, 4, 1, gbl_samples_fp )) == -1 )
	{
	  perror( "xwxapt: fwrite()" );
	  sprintf( message, "Error writing to samples file:\n%s",
		   gbl_samples_file );
	  Show_Message( message, "red" );
          Error_Dialog();
	  return( FALSE );
	}
      if( sample_cnt != 1 )
	{
	  sprintf( message, "Error writing to samples file:\n%s",
		   gbl_samples_file );
	  Show_Message( message, "red" );
	  Error_Dialog();
	  return( FALSE );
	}

      first_call = FALSE;

    } /* if( first_call ) */

  /* Fill samples buffer from DSP */
  Dsp_Fill_Line_Buffer();

  /* Write carrier samples to samples file, abort on error */
  if( (sample_cnt = fwrite(gbl_line_buffer, 1, BLOCK_SIZE,
			   gbl_samples_fp )) == -1 )
    {
      perror( "xwxapt: fwrite()" );
      sprintf( message, "Error writing to samples file:\n%s",
	       gbl_samples_file );
      Show_Message( message, "red" );
      Error_Dialog();
      first_call = TRUE;
      return( FALSE );
    }
  if( (sample_cnt != BLOCK_SIZE) )
    {
      sprintf( message, "Error writing to samples file:\n%s",
	       gbl_samples_file );
      Show_Message( message, "red" );
      Error_Dialog();
      first_call = TRUE;
      return( FALSE );
    }

  /* Stop if line count reaches time */
  /* limit or on user stop request   */
  if( (++line_cnt/2 != gbl_duration) &&
      isFlagSet(ACTION_RECORD_APT) )
    return( TRUE );

  /* Print some information */
  sprintf( message, "Closing samples file:\n%s", gbl_samples_file );
  Show_Message( message, "black" );

  Cleanup();
  first_call = TRUE;

  return( FALSE );

} /* End of Record_APT_Signal() */

/*------------------------------------------------------------------------*/

/*  File_Fill_Buffer()
 *
 *  Fills the audio sample buffer from source file
 */

int
File_Fill_Buffer( void )
{
  /* Count of samples returned by read() */
  int sample_cnt;

  /* Messages buffer */
  char message[350];


  gbl_buffer_idx = 0;

  /* Fill in the samples buffer from file, return -1 on error */
  if( (sample_cnt = fread( gbl_line_buffer, 1,
			   BLOCK_SIZE, gbl_samples_fp )) == -1 )
    {
      perror( "xwxapt: fread()" );
      sprintf( message, "Error reading from samples file:\n%s",
	       gbl_samples_file );
      Show_Message( message, "red" );
      Error_Dialog();
      return(-1);
    }

  if( feof(gbl_samples_fp) ) return( -1 );
  else
    if( sample_cnt != BLOCK_SIZE )
      {
	sprintf( message, "Error reading from samples file:\n%s",
		 gbl_samples_file );
	Show_Message( message, "red" );
	Error_Dialog();
	return(-1);
      }

  /* Set buffer index just after sync train.  */
  /* Search for sync up to +30 from sync ref. */
  Detect_Sync_Train( gbl_sync_refs[gbl_sat_type] + 30 );

  /* If sync detection is far out, keep */
  /* buffer index in default position   */
  if( abs(gbl_buffer_idx - gbl_sync_ref) > 3 )
    gbl_buffer_idx = gbl_sync_ref;

  return(0);

} /* End of File_Fill_Buffer() */

/*------------------------------------------------------------------------*/

/*  Dsp_Fill_Line_Buffer()
 *
 *  Fills the audio sample buffer from sound card DSP.
 *  Buffer filling is synchronized with APT sync train.
 */

int
Dsp_Fill_Line_Buffer( void )
{
  int
    sync_err,   /* Difference between sync ref and pos */
    sync_flag;  /* Sync train detected flag */

  /* Data block size used as argument to read() */
  static int block_size = BLOCK_SIZE;

  gbl_buffer_idx = 0;

  /* Fill in the samples buffer from DSP */
  if( Read_Dsp_Buffer(gbl_line_buffer, block_size+dsp_rate_err ) == -1 )
    {
      Show_Message( "Error filling samples buffer from dsp.", "red" );
      Error_Dialog();
      return(-1);
    }

  /* Set buffer index just after sync train.  */
  /* Search for sync up to +30 from sync ref. */
  Detect_Sync_Train( gbl_sync_refs[gbl_sat_type] + 30 );
  sync_flag = ( gbl_buffer_idx != 0 );

  /* Try to keep APT sync train at beginning of buffer. This  */
  /* done done by increasing or decreasing the data  block    */
  /* size argument passed to read(), by adding or subtracting */
  /* half the difference between detected sync position and a */
  /* reference value. Error is limited to +-1 to avoid coarse */
  /* corrections to sync position due to noise interference.  */
  sync_err = gbl_buffer_idx - gbl_sync_ref;

  if( sync_err > 1 )
    gbl_buffer_idx = gbl_sync_ref + 1;

  if( sync_err < -1 )
    gbl_buffer_idx = gbl_sync_ref - 1;

  /* Vary the data block size argument to read() to   */
  /* take up any drift in sync position due to errors */
  /* in the DSP sampling rate and satellite movement. */
  if( sync_flag ) /* Sync train detected */
    {
      block_size = BLOCK_SIZE + gbl_buffer_idx - gbl_sync_ref;
      if( block_size > BUFFER_SIZE )
	block_size = BUFFER_SIZE;
    }

  return(0);

} /* End of Dsp_Fill_Line_Buffer() */

/*------------------------------------------------------------------------*/

/* Read_Dsp_Buffer()
 *
 * Reads the DSP's buffer and separates Left/Right 
 * channel data depending on stereo or mono mode
 */

int
Read_Dsp_Buffer( unsigned char *buffer, int buff_size )
{
  int
    sample_cnt,
    idx;

  /* Transfer data from dsp buffer to buffer to be processed, */
  /* separating left or right channel data if in stereo mode  */
  for( idx = 0; idx < buff_size; idx++ )
    {
      if( dsp_buf_idx >= chunk_size )
	{
	  /* Fill in the samples buffer from DSP, return -1 on error */
	  if( (sample_cnt = read(gbl_dsp_fd, dsp_buffer, chunk_size)) == -1 )
	      {
		perror( "xwxapt: read()" );
		Show_Message( "Error filling samples buffer from dsp.", "red" );
                Error_Dialog();
		return(-1);
	      }

	  if( sample_cnt != chunk_size )
	    {
	      perror( "xwxapt: read()" );
	      Show_Message( "Error filling samples buffer from dsp.", "red" );
              Error_Dialog();
	      return(-1);
	    }

	  /* Start buffer index according to stereo|mono mode */
	  if( NUM_OF_CHANNELS == 1 ) /* Mono */
	    dsp_buf_idx = 0;
	  else
	    dsp_buf_idx = CHANNEL_IN_USE;

	} /* if( buff_index >= chunk_size ) */

      /* Fill buffer from left or right channel. dsp_buf_idx */
      /* is arranged to point to the correct location */
      buffer[ idx ] = dsp_buffer[ dsp_buf_idx ];

      /* Increment buffer index according to stereo|mono mode */
      dsp_buf_idx += NUM_OF_CHANNELS;

    } /* for( idx = 0; idx < buff_size; idx++ ) */

  return(0);

} /* End of Read_Dsp_Buffer() */

/*------------------------------------------------------------------------*/
