/*  xwxapt.h
 *  
 *  Header file for xwxapt.
 */

/*
 *  xwxapt: An application to decode APT signals from weather 
 *  satellites and produce an image of the weather.
 *
 *  Copyright (C) 2001  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or 
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#ifndef XWXAPT_H
#define XWXAPT_H 1

#include <gtk/gtk.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <stdio.h>
#include <time.h>

/************ CUSTOMIZE THESE DEFINITIONS FOR YOUR NEEDS *************/

#define MIXER_DEV      "/dev/mixer" /* Your mixer device */
#define DSP_DEV          "/dev/dsp" /* Your DSP device   */
#define NUM_OF_CHANNELS           2 /* Stereo/Mono: 1=Mono, 2=Stereo  */
#define CHANNEL_IN_USE         LEFT /* Channel for record: LEFT|RIGHT */
#define RECORD_SRC           "line" /* Sound recording source */
#define RECLEV_DEV          "igain" /* Recording level device */
#define DEFAULT_DURATION        540 /* Default duration (sec) of operations */

/***************** END OF CUSTOMIZABLE SECTION ***********************/

/* Generel definitions for image processing */
#define DSP_SPEED         9600 /* DSP recording speed in samples/sec */
#define RECORD_LEV      0x4040 /* Default level of recording source  */
#define BUFFER_SIZE       4900 /* Audio buffer size=samples/line+100 */
#define BLOCK_SIZE        4800 /* Read() block size=samples/line     */
#define SYNC_DETECT          8 /* Sync-detected threshold (samples)  */
#define MAX_FILE_NAME      255 /* Max length for optional filenames  */
#define BLACK_CUT_OFF        1 /* Black cut-off percentile for normalization */
#define WHITE_CUT_OFF        2 /* White cut-off percentile for normalization */

/* Definitions for NOAA image processing */
#define NOAA_IMAGE_WIDTH          696 /* Width in pixels of images produced */
#define NOAA_SYNC_CHA_SPACE       128 /* Space between ch A sync and image  */
#define NOAA_CHA_CHB_SPACE        310 /* Space between ch A & B, in samples */
#define NOAA_FIFO_SIZE             63 /* Size of sync detector samples fifo */
#define NOAA_CARRIER_MAX          126 /* Max 2.4kHz carrier-level required  */
#define NOAA_SYNC_LOWER_THRESH   1800 /* Sync train detector lower threshold*/
#define NOAA_SYNC_UPPER_THRESH   2400 /* Sync train detector upper threshold*/
#define NOAA_SYNC_REF             150 /* Required position of sync train    */
#define NOAA_SCALE_FACTOR          90 /* Factor to keep pixel values < 255  */

/* Definitions for Meteor image processing */
#define METEOR_IMAGE_WIDTH        768 /* Width in pixels of images produced */
#define METEOR_SYNC_IMAGE_SPACE   130 /* Space between sync and image start */
#define METEOR_FIFO_SIZE           40 /* Size of sync detector samples fifo */
#define METEOR_CARRIER_MAX        160 /* Max 2.4kHz carrier-level required  */
#define METEOR_SYNC_LOWER_THRESH 1500 /* Sync train detector lower threshold*/
#define METEOR_SYNC_UPPER_THRESH 2200 /* Sync train detector upper threshold*/
#define METEOR_SYNC_REF           130 /* Required position of sync train    */
#define METEOR_SCALE_FACTOR        60 /* Factor to keep pixel values < 255  */

/* Satellite type definitions */
#define NOAA                      0 /* NOAA type APT format   */
#define METEOR                    1 /* METEOR type APT format */
#define NUM_SATS                  2 /* Number of APT formats supported */
#define SAT_NAMES  "NOAA", "Meteor"
#define SYNC_REFS   NOAA_SYNC_REF, METEOR_SYNC_REF

/* Channel to be used in audio capture */
#define LEFT                      0 /* Left channel  */
#define RIGHT                     1 /* Right channel */

/* Flow control flags */
#define ICON_SYNC_LOCK          0x0001 /* Sync detector locked icon    */
#define ICON_SYNC_UNLOCK        0x0002 /* Sync detector unlocked icon  */
#define ICON_SYNC_SEARCH        0x0004 /* Sync detector searching icon */
#define ICON_SYNC_STOPPED       0x0008 /* Sync detector searching icon */
#define ICON_FLAGS_ALL          0x000f /* All icon flags (for clearing)*/
#define ACTION_PROCESS_DSP      0x0010 /* Process images from dsp      */
#define ACTION_PROCESS_FILE     0x0020 /* Process images from file     */
#define ACTION_RECORD_APT       0x0040 /* Record apt signal to file    */
#define ACTION_SETUP_AUDIO      0x0080 /* Run audio level setup func.  */
#define ACTION_SYNC_BUFFER      0x0100 /* Run audio level setup func.  */
#define ACTION_FLAGS_ALL        0x01f0 /* All action flags (clearing)  */
#define IMAGE_NORMALIZE         0x0400 /* Histogram normalize wx image */
#define IMAGE_ROTATE            0x0800 /* Rotate wx image 180 degrees  */
#define SOUND_CARD_SETUP        0x1000 /* Sound card has been set up   */
#define START_STOP_TIMER        0x2000 /* Start-Stop timer is set      */

/*---------------------------------------------------------------------*/

/* Function prototypes */

/* image.c */
unsigned
char     Scalar_Mag(unsigned char x, unsigned char y);
int      Detect_Sync_Train(int search_limit);
int      Detect_NOAA_Sync(int search_limit);
int      Detect_Meteor_Sync(int search_limit);
int      NOAA_Image_Line(unsigned char *image_line, int *max_gray_val);
int      Meteor_Image_Line(unsigned char *image_line, int *max_gray_val);
gboolean Dsp_Process_Image(gpointer data);
gboolean File_Process_Image(gpointer data);
void     Normalize(unsigned char *image_buffer, int image_size);
void     Rotate(unsigned char *image_buffer, int image_size);
void     Set_Sync_Icon(gchar *img_name);
/* main.c */
int      main(int argc, char *argv[]);
void     Usage(void);
void     File_Name(char *file_name);
void     Cleanup(void);
void     Show_Message( char *message, char *attr);
int      isFlagSet(int flag);
int      isFlagClear(int flag);
void     SetFlag(int flag);
void     ClearFlag(int flag);
void     ToggleFlag(int flag);
void     Initialize_Top_window(void);
/* sound.c */
void     Setup_Sound_Card(void);
gboolean Setup_Audio_Level(gpointer data);
gboolean Synchronize_Buffer(gpointer data);
gboolean Record_APT_Signal(gpointer data);
int      File_Fill_Buffer(void);
gboolean Dsp_Fill_Line_Buffer(void);
int      Read_Dsp_Buffer(unsigned char *buffer, int block_size);
/* callbacks.c */
void     Error_Dialog(void);
void     File_Selection(void);
void     Sensitize_Menu_Item(gchar *item_name, gboolean flag);
void     Popup_Menu(void);

/*-------------------------------------------------------------------*/

#endif /* End of xwxapt.h */
